# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from muranoclient.common import exceptions as exc
from muranoclient.v1 import client
from muranodashboard.common import utils
from muranodashboard.environments import api as env_api
from muranodashboard.environments import consts
from openstack_dashboard.test import helpers


class TestEnvironmentsAPI(helpers.APITestCase):
    def setUp(self):
        super(TestEnvironmentsAPI, self).setUp()

        self.mock_client = mock.Mock(spec=client)
        self.mock_request = mock.MagicMock()
        self.mock_request.session = {'django_timezone': 'UTC'}
        self.env_id = 'foo_env_id'
        self.session_id = 'foo_session_id'
        self.service_id = 'foo_service_id'
        self.deployment_id = 'foo_deployment_id'

        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_get_status_messages_for_service(self, mock_log, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.deployments.list.return_value = [
            mock.Mock(id='foo_deployment_id'),
            mock.Mock(id='bar_deployment_id')
        ]

        mock_client.deployments.reports.side_effect = [
            [mock.Mock(text='foo_text', created='1970-01-01T12:23:00')],
            [mock.Mock(text='bar_text', created='1970-01-01T15:45:00')],
        ]

        expected_result = '\n1970-01-01 12:23:00 - foo_text\n' \
                          '1970-01-01 15:45:00 - bar_text\n'
        expected_reports_mock_calls = [
            mock.call('foo_env_id', 'bar_deployment_id', 'foo_service_id'),
            mock.call('foo_env_id', 'foo_deployment_id', 'foo_service_id')
        ]

        result = env_api.get_status_messages_for_service(
            self.mock_request, self.service_id, self.env_id)

        self.assertEqual(expected_result, result)
        mock_client.deployments.reports.assert_has_calls(
            expected_reports_mock_calls)
        mock_client.deployments.list.assert_called_once_with('foo_env_id')
        self.assertTrue(mock_log.debug.called)

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_environment_update(self, mock_api):
        env_name = "test_env"
        env_api.environment_update(self.mock_request, self.env_id, env_name)
        env_api.api.muranoclient.assert_called_once_with(self.mock_request)

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_environment_list(self, mock_api):
        env_api.environments_list(self.mock_request)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        env_api.api.handled_exceptions.assert_called_with(self.mock_request)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_environment_create(self, mock_log, mock_api):
        parameters = {
            'name': 'test_env',
            'defaultNetworks': 'test_net'
        }
        env_api.environment_create(self.mock_request, parameters)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        self.assertTrue(mock_log.debug.called)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_environment_delete(self, mock_log, mock_api):
        env_api.environment_delete(self.mock_request, self.env_id)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        (mock_log.debug.
            assert_called_once_with('Environment::{0} <Id :'
                                    ' {1}>'.format('Delete', self.env_id)))

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_environment_deploy(self, mock_log, mock_api):
        env_api.environment_deploy(self.mock_request, self.env_id)
        self.assertTrue(env_api.api.muranoclient.called)
        self.assertTrue(mock_log.debug.called)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_action_allowed(self, mock_log, mock_api):
        result = env_api.action_allowed(self.mock_request, self.env_id)
        self.assertTrue(result)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        self.assertTrue(mock_log.debug.called)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_service_create(self, mock_log, mock_api):
        parameters = {
            '?': {
                'type': 'test.Service'
            }
        }
        env_api.service_create(self.mock_request, self.env_id, parameters)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        mock_log.debug.assert_called_with('Service::Create {0}'
                                          .format(parameters['?']['type']))

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'Session', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_service_delete(self, mock_log, mock_session, mock_api):
        mock_response = mock.Mock(status_code=200)
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.services.delete.return_value = mock_response
        mock_session.get_or_create_or_delete.return_value = self.session_id

        result = env_api.service_delete(self.mock_request, self.env_id,
                                        self.service_id)

        self.assertEqual(mock_response, result)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        mock_session.get_or_create_or_delete.assert_called_with(
            self.mock_request, 'foo_env_id')
        mock_client.services.delete.assert_called_once_with(
            'foo_env_id', '/foo_service_id', 'foo_session_id')
        mock_log.debug.assert_called_with(
            'Service::Delete <SrvId: {0}>'.format('foo_service_id'))

    @mock.patch.object(env_api, 'services_list', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_service_get(self, mock_log, mock_services_list):
        mock_services_list.return_value = [{'?': {'id': 'foo_service_id'}}]
        result = env_api.service_get(self.mock_request, self.env_id,
                                     'foo_service_id')

        self.assertEqual({'?': {'id': 'foo_service_id'}}, result)
        mock_services_list.assert_called_once_with(
            self.mock_request, 'foo_env_id')
        mock_log.debug.assert_called_with(
            'Return service detail for a specified id')

    def test_extract_actions_list(self):
        service = {
            '?': {
                'test': 'test'
            }
        }
        result = env_api.extract_actions_list(service)
        self.assertEqual([], result)

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_run_action(self, mock_api):
        env_api.run_action(self.mock_request, self.env_id, 'foo_action_id')
        env_api.api.muranoclient.assert_called_with(self.mock_request)

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_deployment_reports(self, mock_log, mock_api):
        env_api.deployment_reports(self.mock_request, self.env_id,
                                   self.deployment_id)
        env_api.api.muranoclient.assert_called_with(self.mock_request)
        self.assertTrue(mock_log.debug.called)

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_deployment_start(self, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.deployments.list.return_value = []
        result = env_api.get_deployment_start(self.mock_request, self.env_id,
                                              self.deployment_id)
        self.assertIsNone(result)

        mock_client.deployments.list.return_value = [
            mock.Mock(id='foo_deployment_id', started='1970-01-01T12:34:00')
        ]
        result = env_api.get_deployment_start(self.mock_request, self.env_id,
                                              self.deployment_id)
        self.assertEqual('1970-01-01 12:34:00', result)
        mock_client.deployments.list.assert_has_calls([
            mock.call('foo_env_id'), mock.call('foo_env_id')
        ])

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_deployment_description(self, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.deployments.list.return_value = []
        result = env_api.get_deployment_start(self.mock_request, self.env_id,
                                              self.deployment_id)
        self.assertIsNone(result)

        mock_client.deployments.list.return_value = [
            mock.Mock(id='foo_deployment_id', description='foo_descr')
        ]
        result = env_api.get_deployment_descr(self.mock_request, self.env_id,
                                              self.deployment_id)
        self.assertEqual('foo_descr', result)
        mock_client.deployments.list.assert_has_calls([
            mock.call('foo_env_id'), mock.call('foo_env_id')
        ])

    @mock.patch('muranodashboard.environments.api.topology.json.dumps',
                autospec=True)
    @mock.patch('muranodashboard.environments.api.topology._environment_info',
                autospec=True)
    @mock.patch.object(env_api, 'environment_get', autospec=True)
    def test_load_environment_data(self, mock_env_get,
                                   mock_env_info, mock_dump):
        mock_env_info.return_value = 'services/_environment_info.html'
        mock_dump.return_value = self.env_id
        result = env_api.load_environment_data(self.mock_request, self.env_id)
        self.assertTrue(mock_env_get.called)
        self.assertEqual(self.env_id, result)
        self.assertTrue(mock_dump.called)

    @mock.patch.object(env_api, 'deployments_list', autospec=True)
    def test_update_env_return_ready_status(self, mock_deployments_list):
        mock_deployments_list.return_value = []
        mock_env = mock.Mock(id='foo_env_id', services=[], version=1,
                             status=consts.STATUS_ID_PENDING)

        result = env_api._update_env(mock_env, self.mock_request)
        self.assertEqual(mock_env, result)
        self.assertEqual(consts.STATUS_ID_READY, result.status)
        self.assertFalse(result.has_new_services)

    @mock.patch.object(env_api, 'deployments_list', autospec=True)
    def test_update_env_return_new_status(self, mock_deployments_list):
        mock_deployments_list.return_value = []
        mock_env = mock.Mock(id='foo_env_id', services=[], version=0,
                             status=consts.STATUS_ID_READY)

        result = env_api._update_env(mock_env, self.mock_request)
        self.assertEqual(mock_env, result)
        self.assertEqual(consts.STATUS_ID_NEW, result.status)
        self.assertFalse(result.has_new_services)

    @mock.patch.object(env_api, 'Session', autospec=True)
    @mock.patch.object(env_api, 'packages_api', autospec=True)
    @mock.patch.object(env_api, 'api', autospec=True)
    def test_services_list(self, mock_api, mock_pkg_api, mock_session):
        mock_env = mock.Mock(version=0)
        mock_env.services = [
            {'?': {'id': 'foo_service_id', 'name': 'foo', 'type': 'foo_type'}},
            {'?': {'id': 'bar_service_id', 'name': 'bar',
                   'type': '/3@bar_type'}, 'updated': 'bar_time'},
        ]
        mock_foo_pkg = mock.Mock()
        mock_foo_pkg.configure_mock(name='foo_pkg')
        mock_bar_pkg = mock.Mock()
        mock_bar_pkg.configure_mock(name='bar_pkg')

        mock_session.get.return_value = 'foo_sess_id'
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.environments.get.return_value = mock_env
        mock_client.environments.last_status.return_value = {
            'foo_service': mock.Mock(text='foo' * 100,
                                     updated='foo_time'),
            'bar_service': None
        }
        mock_pkg_api.app_by_fqn.side_effect = [mock_foo_pkg, mock_bar_pkg]

        expected_pkg_calls = [
            mock.call(self.mock_request, 'foo_type', version=None),
            mock.call(self.mock_request, 'bar_type', version='3')
        ]

        result = env_api.services_list(self.mock_request, 'foo_env_id')

        self.assertIsInstance(result, list)
        for obj in result:
            self.assertIsInstance(obj, utils.Bunch)

        mock_session.get.assert_called_once_with(
            self.mock_request, 'foo_env_id')
        mock_client.environments.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        mock_client.environments.last_status.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        mock_pkg_api.app_by_fqn.assert_has_calls(expected_pkg_calls)

    @mock.patch.object(env_api, 'LOG', autospec=True)
    @mock.patch.object(env_api, 'Session', autospec=True)
    @mock.patch.object(env_api, 'packages_api', autospec=True)
    @mock.patch.object(env_api, 'api', autospec=True)
    def test_services_list_except_http_not_found(self, mock_api, mock_pkg_api,
                                                 mock_session, mock_log):
        mock_env = mock.Mock(version=0)
        mock_env.services = [
            {'?': {'id': 'foo_service_id', 'name': 'foo', 'type': 'foo_type'}}
        ]
        mock_foo_pkg = mock.Mock()
        mock_foo_pkg.configure_mock(name='foo_pkg')

        mock_session.get.return_value = 'foo_sess_id'
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.environments.get.return_value = mock_env
        mock_client.environments.last_status.side_effect = exc.HTTPNotFound
        mock_pkg_api.app_by_fqn.side_effect = [mock_foo_pkg]

        expected_pkg_calls = [
            mock.call(self.mock_request, 'foo_type', version=None)
        ]

        result = env_api.services_list(self.mock_request, 'foo_env_id')

        self.assertIsInstance(result, list)
        for obj in result:
            self.assertIsInstance(obj, utils.Bunch)

        mock_session.get.assert_called_once_with(
            self.mock_request, 'foo_env_id')
        mock_client.environments.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        mock_client.environments.last_status.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        mock_log.exception.assert_called_once_with(
            'Could not retrieve latest status for the foo_env_id environment')
        mock_pkg_api.app_by_fqn.assert_has_calls(expected_pkg_calls)

    @mock.patch.object(env_api, 'services_list', autospec=True)
    def test_service_list_by_fqns(self, mock_services_list):
        self.assertEqual([], env_api.service_list_by_fqns(None, None, []))

        mock_services_list.return_value = [
            {'?': {'type': 'foo/bar'}}, {'?': {'type': 'baz/qux'}}
        ]
        result = env_api.service_list_by_fqns(
            self.mock_request, 'foo_env_id', ['foo'])
        self.assertEqual([{'?': {'type': 'foo/bar'}}], result)


class TestEnvironmentsSessionAPI(helpers.APITestCase):
    def setUp(self):
        super(TestEnvironmentsSessionAPI, self).setUp()

        self.mock_client = mock.Mock(spec=client)
        self.mock_request = mock.MagicMock()
        self.env_id = 'foo_env_id'

        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_or_create(self, mock_api):
        self.session = env_api.Session()
        result = self.session.get_or_create(self.mock_request, self.env_id)
        self.assertIsNotNone(result)
        env_api.api.muranoclient.assert_called_once_with(self.mock_request)

    def test_set(self):
        session_id = 11
        self.session = env_api.Session()
        result = self.session.set(self.mock_request, self.env_id, session_id)
        self.assertIsNone(result)

    @mock.patch.object(env_api, 'create_session', autospec=True)
    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_gcd(self, mock_log, mock_api, mock_create_session):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_session_data = mock.Mock(state=consts.STATUS_ID_READY)
        mock_client.sessions.get.return_value = mock_session_data
        mock_create_session.return_value = 'bar_sess_id'

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_or_create_or_delete(self.mock_request,
                                                         'foo_env_id')

        self.assertEqual('bar_sess_id', result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        self.assertNotIn('foo_env_id', self.mock_request.session)
        mock_log.debug.assert_called_once_with(
            'The existing session has been already deployed. Creating a new '
            'session for the environment foo_env_id')
        mock_create_session.assert_called_once_with(
            self.mock_request, 'foo_env_id')

    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_gcd_with_active_session(self, mock_log, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_session_data = mock.Mock(state='foo_bar_state')
        mock_client.sessions.get.return_value = mock_session_data

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_or_create_or_delete(self.mock_request,
                                                         'foo_env_id')

        self.assertEqual('foo_sess_id', result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        self.assertNotIn('foo_env_id', self.mock_request.session)
        mock_log.debug.assert_called_once_with(
            'Found active session for the environment foo_env_id')

    @mock.patch.object(env_api, 'create_session', autospec=True)
    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_gcd_with_new_session(self, mock_log, mock_api,
                                  mock_create_session):
        mock_create_session.return_value = 'bar_sess_id'

        result = env_api.Session.get_or_create_or_delete(self.mock_request,
                                                         'foo_env_id')

        self.assertEqual('bar_sess_id', result)
        mock_log.debug.assert_called_once_with('Creating a new session')
        mock_create_session.assert_called_once_with(
            self.mock_request, 'foo_env_id')

    @mock.patch.object(env_api, 'create_session', autospec=True)
    @mock.patch.object(env_api, 'api', autospec=True)
    @mock.patch.object(env_api, 'LOG', autospec=True)
    def test_gcd_except_http_forbidden(self, mock_log, mock_api,
                                       mock_create_session):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.sessions.get.side_effect = exc.HTTPForbidden
        mock_create_session.return_value = 'bar_sess_id'

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_or_create_or_delete(self.mock_request,
                                                         'foo_env_id')

        self.assertEqual('bar_sess_id', result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')
        self.assertNotIn('foo_env_id', self.mock_request.session)
        mock_log.debug.assert_called_once_with(
            'The environment is being deployed by other user. '
            'Creating a new session for the environment foo_env_id')
        mock_create_session.assert_called_once_with(
            self.mock_request, 'foo_env_id')

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_if_available(self, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.sessions.get.return_value = mock.Mock(
            state='foo_bar_state')

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_if_available(
            self.mock_request, 'foo_env_id')

        self.assertEqual('foo_sess_id', result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_if_available_with_none_returned(self, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.sessions.get.return_value = \
            mock.Mock(state=consts.STATUS_ID_READY)

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_if_available(
            self.mock_request, 'foo_env_id')

        self.assertIsNone(result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')

    @mock.patch.object(env_api, 'api', autospec=True)
    def test_get_if_available_except_http_forbidden(self, mock_api):
        mock_client = mock_api.muranoclient(mock.Mock())
        mock_client.sessions.get.side_effect = exc.HTTPForbidden

        self.mock_request.session = {'sessions': {'foo_env_id': 'foo_sess_id'}}
        result = env_api.Session.get_if_available(
            self.mock_request, 'foo_env_id')

        self.assertIsNone(result)
        mock_client.sessions.get.assert_called_once_with(
            'foo_env_id', 'foo_sess_id')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

PackageDir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                          'MockApp')
HotPackageDir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             'HotApp')
DeployingPackageDir = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                   'DeployingApp')

CategorySelector = "//a[contains(text(), '{0}')][contains(@class, 'dropdown-toggle')]"  # noqa
EnvAppsCategorySelector = "//*[contains(@id, 'envAppsCategoryBtn')]"
App = "//div[contains(@class, 'app-list')]//h4[contains(text(), '{0}')]"
Component = "//div[contains(@id, apps_carousel)]//div[contains(text(), '{0}')]"
MockAppDescr = "//div[h4[contains(text(), 'MockApp')]]/p"
AppPackages = u"//tr[@data-display='{0}']"
AppDetail = "//dl[dt[contains(text(), 'Name')]]/dd"
TagInDetails = "//div[contains(@class, 'app-meta')]//ul//li[strong[contains(text(), 'Tags')]]"  # noqa
TestImage = "//tr[td[contains(text(), '{0}')]]"
DeleteImageMeta = TestImage + "//td//button[contains(text(), 'Delete Metadata')]"  # noqa
ImageMeta = "//dl//div[dt[contains(text(), 'murano_image_info')]]/dd"
More = "//tr[contains(@id, '{0}__row__{1}')]//a[contains(@class, dropdown-toggle) and @href='#']"  # noqa
Status = "//td[contains(text(), '{0}')]"
EnvStatus = "//tr[contains(@data-display, '{0}')]/td[contains(text(), '{1}')]"
CellStatus = "//td[contains(@class, 'status_{0}')]"
Row = "//tr[contains(@id, 'services__row__{0}')]"
ErrorMessage = '//span[contains(@class, "help-block") and contains(text(), "{0}")]'  # noqa
JsErrorMessage = '//div[contains(@class, "alert-danger") and contains(text(), "{0}")]'  # noqa
EnvAppsCategory = "//div[contains(@class, 'draggable_app')]//div[contains(text(), '{0}')]"  # noqa
PackageCategory = "//select[@name='add_category-categories']/option[text()='{0}']"  # noqa
DatabaseCategory = "select[name='add_category-categories'] > option[value='Databases']"  # noqa
CategoryPackageCount = "//tr[contains(@data-display, '{0}')]/td[contains(text(), '{1}')]"  # noqa
Action = '//a[contains(@class, "murano_action") and contains(text(), "testAction")]'  # noqa
HotFlavorField = '//div[contains(@class, "has-error")]//input'
EnvCheckbox = "//tr[contains(@data-display, '{0}')]/td[contains(@class, 'multi_select_column')]//div//label"  # noqa
NewEnvRow = "table#environments thead tr.new_env"
TableSorterByName = "table#environments thead th.tablesorter-header[data-column='1']"  # noqa
ServiceDetail = "//dd[contains(text(), '{0}')]"
ServiceType = "//table[@id='services']//tbody//tr//td[2][contains(text(), '{0}')]"  # noqa
TableDropdownBtn = "//tr[contains(@data-display, '{0}')]//a[contains(@class, "\
                   "'dropdown-toggle')]"
TableDropdownMenu = "//tr[contains(@data-display, '{0}')]//div[contains("\
                    "@class, 'open')]"
TableDropdownAction = "//tr[contains(@data-display, '{0}')]//button[contains("\
                      "text(), '{1}')]"

# Buttons
ButtonSubmit = ".//*[@name='wizard_goto_step'][2]"
InputSubmit = "//input[@type='submit']"
NextWizardSubmit = 'div.modal-footer input[value="Next"]'
CreateWizardSubmit = 'div.modal-footer input[value="Create"]'
ConfirmDeletion = "//div[@class='modal-footer']//a[contains(text(), 'Delete')]"  # noqa
ConfirmAbandon = "//div[@class='modal-footer']//a[contains(text(), 'Abandon')]"  # noqa
UploadPackage = 'packages__action_upload_package'
ImportBundle = 'packages__action_import_bundle'
CreateEnvironment = ".add_env .btn"
DeployEnvironment = "services__action_deploy_env"
DeleteEnvironment = "//button[contains(@id, 'action_delete')]"
DeployEnvironments = ".btn#environments__action_deploy"
DeployEnvironmentsDisabled = ".btn#environments__action_deploy[disabled]"
DeleteEnvironments = ".btn#environments__action_delete"
DeleteEnvironmentsDisabled = ".btn#environments__action_delete[disabled]"
AbandonEnvironment = "//button[contains(text(), 'Abandon Environment')]"
AbandonEnvironments = ".btn#environments__action_abandon"
AbandonEnvironmentsDisabled = ".btn#environments__action_abandon[disabled]"
ConfirmCreateEnvironment = 'confirm_create_env'
MoreActions = ".table_actions_menu > .btn"
AddComponent = "services__action_AddApplication"
AddCategory = "categories__action_add_category"
DeleteCategory = "//tr[td[contains(text(), '{0}')]]//button[contains(@id, 'action_delete')]"  # noqa
NextBtn = "//tfoot//tr//td//a[contains(@href,'?marker')]"
PrevBtn = "//tfoot//tr//td//a[contains(@href,'prev_marker')]"
DeleteComponent = ".btn[id^='services__row_'][id$='__action_delete']"
DetailDropdownBtn = "form.detail-actions-form a.dropdown-toggle"
DetailDropdownMenu = "ul.dropdown-menu"
DeploymentHistoryLogTab = "//ul[contains(@id, 'environment_details')]//"\
                          "a[contains(text(), 'Latest Deployment Log')]"
EnvComponentsTab = "//ul[contains(@id, 'environment_details')]//"\
                   "a[contains(text(), 'Components')]"
DeploymentHistoryLogs = "div#environment_details__env_logs div.reports.logs "\
                        "div.report-info"
PackageFilterDropdownBtn = 'div.table_search > div.themable-select.dropdown >'\
                           ' button'
PackageFilterTypeBtn = "a[data-select-value='{0}']"
PackageFilterInput = 'input[name="packages__filter_packages__q"]'
PackageFilterBtn = "packages__action_filter_packages"

# Panels
AppCatalog = "//*[@id='main_content']/nav//a[contains(text(), 'App Catalog')]"  # noqa
Browse = AppCatalog + "/following::a[contains(text(), 'Browse')]"
Manage = AppCatalog + "/following::a[contains(text(), 'Manage')]"
Applications = AppCatalog + "/following::a[contains(text(), 'Applications')]"  # noqa
AlertInfo = "//*[contains(@class, 'alert-info')][contains(text(), '{0}')]"

# Modals
ModalDialog = ".modal-dialog"



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

WF_MANAGEMENT_NAME = 'workflowManagement'


class WorkflowManagementForm(object):
    def __init__(self):
        self.name = WF_MANAGEMENT_NAME
        self.field_specs = [
            {'name': 'stay_at_the_catalog',
             'initial': False,
             'description': 'If checked, you will be returned to the '
                            'Catalog page. If not - to the '
                            'Environment page, where you can deploy'
                            ' the application.',
             'required': False,
             'type': 'boolean',
             'label': 'Continue application adding'}]
        self.validators = []

    def name_field(self, fqn):
        return {'name': 'application_name',
                'type': 'string',
                'description': 'Enter a desired name for the application. '
                               'Just A-Z, a-z, 0-9, dash and underline'
                               ' are allowed',
                'label': 'Application Name',
                'regexpValidator': r'^[-\w]+$',
                'initial': fqn.split('.')[-1]
                }



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from designateclient import client

from keystoneauth1 import loading
from keystoneauth1 import session

from openstack_dashboard.api import base


def designateclient(request, api_version='2'):
    loader = loading.get_plugin_loader('token')
    auth = loader.load_from_options(
        auth_url=base.url_for(request, 'identity'),
        token=request.user.token.id,
        project_id=request.user.project_id,
        project_domain_id=request.user.token.project.get('domain_id'))

    sess = session.Session(auth=auth)
    return client.Client(api_version, session=sess)


def zone_list(request):
    d_client = designateclient(request)
    if d_client is None:
        return []
    return d_client.zones.list()



# The name of the panel to be added to HORIZON_CONFIG. Required.
PANEL = 'images'
# The name of the dashboard the PANEL associated with. Required.
PANEL_DASHBOARD = 'app-catalog'
# The name of the panel group the PANEL is associated with.
PANEL_GROUP = 'app-catalog_manage_group'

# Python panel class of the PANEL to be added.
ADD_PANEL = 'muranodashboard.images.panel.Images'



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from django import http as django_http
from django import template
from django.template import defaultfilters
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy

from horizon import exceptions
from horizon import forms
from horizon import messages
from horizon import tables
from horizon.utils import filters
from muranoclient.common import exceptions as exc
from openstack_dashboard import policy
from oslo_log import log as logging

from muranodashboard import api as api_utils
from muranodashboard.api import packages as pkg_api
from muranodashboard.catalog import views as catalog_views
from muranodashboard.common import utils as md_utils
from muranodashboard.environments import api
from muranodashboard.environments import consts
from muranodashboard.packages import consts as pkg_consts

LOG = logging.getLogger(__name__)


def _get_environment_status_and_version(request, table):
    environment_id = table.kwargs.get('environment_id')
    env = api.environment_get(request, environment_id)
    status = getattr(env, 'status', None)
    version = getattr(env, 'version', None)
    return status, version


def _check_row_actions_allowed(action, request):
    envs = action.table.data
    if not envs:
        return False
    for env in envs:
        if action.allowed(request, env):
            return True
    return False


def _environment_has_deployed_services(request, environment_id):
    deployments = api.deployments_list(request, environment_id)
    if not deployments:
        return False
    if not deployments[0].description['services']:
        return False
    return True


class AddApplication(tables.LinkAction):
    name = 'AddApplication'
    verbose_name = _('Add Component')
    icon = 'plus'

    def allowed(self, request, environment):
        status, version = _get_environment_status_and_version(request,
                                                              self.table)
        return status not in consts.NO_ACTION_ALLOWED_STATUSES

    def get_link_url(self, datum=None):
        base_url = reverse('horizon:app-catalog:catalog:switch_env',
                           args=(self.table.kwargs['environment_id'],))
        redirect_url = reverse('horizon:app-catalog:catalog:index')
        return '{0}?next={1}'.format(base_url, redirect_url)


class CreateEnvironment(tables.LinkAction):
    name = 'CreateEnvironment'
    verbose_name = _('Create Environment')
    url = 'horizon:app-catalog:environments:create_environment'
    classes = ('btn-launch', 'add_env')
    redirect_url = "horizon:app-catalog:environments:index"
    icon = 'plus'
    policy_rules = (("murano", "create_environment"),)

    def allowed(self, request, datum):
        return True if self.table.data else False

    def action(self, request, environment):
        try:
            api.environment_create(request, environment)
        except Exception:
            msg = _('Unable to create environment "%s".') % environment
            LOG.error(msg)
            redirect = reverse(self.redirect_url)
            exceptions.handle(request, msg, redirect=redirect)


class DeploymentHistory(tables.LinkAction):
    name = 'DeploymentHistory'
    verbose_name = _('Deployment History')
    url = 'horizon:app-catalog:environments:deployment_history'
    classes = ('deployment-history')
    redirect_url = "horizon:app-catalog:environments:index"
    icon = 'history'
    policy_rules = (("murano", "list_deployments_all_environments"),)

    def allowed(self, request, datum):
        return True


class DeleteEnvironment(policy.PolicyTargetMixin, tables.DeleteAction):
    redirect_url = "horizon:app-catalog:environments:index"
    policy_rules = (("murano", "delete_environment"),)

    @staticmethod
    def action_present(count):
        return ngettext_lazy(
            u"Delete Environment",
            u"Delete Environments",
            count
        )

    @staticmethod
    def action_past(count):
        return ngettext_lazy(
            u"Started Deleting Environment",
            u"Started Deleting Environments",
            count
        )

    def allowed(self, request, environment):
        # table action case: action allowed if any row action allowed
        if not environment:
            return _check_row_actions_allowed(self, request)

        # row action case
        return environment.status not in (consts.STATUS_ID_DEPLOYING,
                                          consts.STATUS_ID_DELETING)

    def action(self, request, environment_id):
        try:
            api.environment_delete(request, environment_id)
        except Exception:
            msg = _('Unable to delete environment "%s".') % environment_id
            LOG.error(msg)
            redirect = reverse(self.redirect_url)
            exceptions.handle(request, msg, redirect=redirect)


class AbandonEnvironment(tables.DeleteAction):
    help_text = _("This action cannot be undone. Any resources created by "
                  "this environment will have to be released manually.")
    name = 'abandon'
    redirect_url = "horizon:app-catalog:environments:index"
    policy_rules = (("murano", "delete_environment"),)

    def __init__(self, **kwargs):
        super(AbandonEnvironment, self).__init__(**kwargs)
        self.icon = 'stop'

    @staticmethod
    def action_present(count):
        return ngettext_lazy(
            u"Abandon Environment",
            u"Abandon Environments",
            count
        )

    @staticmethod
    def action_past(count):
        return ngettext_lazy(
            u"Abandoned Environment",
            u"Abandoned Environments",
            count
        )

    def allowed(self, request, environment):
        """Limit when 'Abandon Environment' button is shown

        'Abandon Environment' button is hidden in several cases:
         * environment is new
         * app added to env, but not deploy is not started
        """

        # table action case: action allowed if any row action allowed
        if not environment:
            return _check_row_actions_allowed(self, request)

        # row action case
        status = getattr(environment, 'status', None)
        if status in [consts.STATUS_ID_NEW, consts.STATUS_ID_PENDING]:
            return False
        return True

    def action(self, request, environment_id):
        try:
            api.environment_delete(request, environment_id, True)
        except Exception:
            msg = _('Unable to abandon the environment "%s".') % environment_id
            LOG.error(msg)
            redirect = reverse(self.redirect_url)
            exceptions.handle(request, msg, redirect=redirect)


class DeleteService(tables.DeleteAction):

    @staticmethod
    def action_present(count):
        return ngettext_lazy(
            u"Delete Component",
            u"Delete Components",
            count
        )

    @staticmethod
    def action_past(count):
        return ngettext_lazy(
            u"Started Deleting Component",
            u"Started Deleting Components",
            count
        )

    def allowed(self, request, service=None):
        status, version = _get_environment_status_and_version(request,
                                                              self.table)
        return status != consts.STATUS_ID_DEPLOYING

    def action(self, request, service_id):
        try:
            environment_id = self.table.kwargs.get('environment_id')
            for service in self.table.data:
                if service['?']['id'] == service_id:
                    api.service_delete(request,
                                       environment_id,
                                       service_id)
        except Exception:
            msg = _('Sorry, you can\'t delete service right now')
            redirect = reverse("horizon:app-catalog:environments:index")
            exceptions.handle(request, msg, redirect=redirect)


class DeployEnvironment(tables.BatchAction):
    name = 'deploy'
    classes = ('btn-launch',)
    icon = "play"

    @staticmethod
    def action_present_deploy(count):
        return ngettext_lazy(
            u"Deploy Environment",
            u"Deploy Environments",
            count
        )

    @staticmethod
    def action_past_deploy(count):
        return ngettext_lazy(
            u"Started deploying Environment",
            u"Started deploying Environments",
            count
        )

    @staticmethod
    def action_present_update(count):
        return ngettext_lazy(
            u"Update Environment",
            # there can be cases when some of the envs are new and some are not
            # so it is better to just leave "Deploy" for multiple envs
            u"Deploy Environments",
            count
        )

    @staticmethod
    def action_past_update(count):
        return ngettext_lazy(
            u"Updated Environment",
            u"Deployed Environments",
            count
        )

    action_present = action_present_deploy
    action_past = action_past_deploy

    def allowed(self, request, environment):
        """Limit when 'Deploy Environment' button is shown

        'Deploy environment' is shown when set of environment's services
        changed or previous deploy failed.
        If environment has already deployed services,
        button is shown as 'Update environment'
        """

        # table action case: action allowed if any row action allowed
        if not environment:
            return _check_row_actions_allowed(self, request)

        # row action case
        if _environment_has_deployed_services(request, environment.id):
            self.action_present = self.action_present_update
            self.action_past = self.action_past_update
        else:
            self.action_present = self.action_present_deploy
            self.action_past = self.action_past_deploy

        status = getattr(environment, 'status', None)
        if status in (consts.STATUS_ID_PENDING,
                      consts.STATUS_ID_DEPLOY_FAILURE):
            return True
        return False

    def action(self, request, environment_id):
        try:
            api.environment_deploy(request, environment_id)
        except Exception:
            msg = _('Unable to deploy. Try again later')
            redirect = reverse('horizon:app-catalog:environments:index')
            exceptions.handle(request, msg, redirect=redirect)


class DeployThisEnvironment(tables.Action):
    name = 'deploy_env'
    verbose_name = _('Deploy This Environment')
    requires_input = False
    classes = ('btn-launch',)
    icon = "play"

    def allowed(self, request, service):
        """Limit when 'Deploy This Environment' button is shown

        'Deploy environment' is not shown in several cases:
        * when deploy is already in progress
        * delete is in progress
        * env was just created and no apps added
        * previous deployment finished successfully
        If environment has already deployed services, button is shown
        as 'Update This Environment'
        """
        environment_id = self.table.kwargs['environment_id']
        if _environment_has_deployed_services(request, environment_id):
            self.verbose_name = _('Update This Environment')
        else:
            self.verbose_name = _('Deploy This Environment')

        status, version = _get_environment_status_and_version(request,
                                                              self.table)
        if (status in consts.NO_ACTION_ALLOWED_STATUSES or
                status == consts.STATUS_ID_READY):
            return False

        apps = self.table.data
        if version == 0 and not apps:
            return False
        return True

    def single(self, data_table, request, service_id):
        environment_id = data_table.kwargs['environment_id']
        try:
            api.environment_deploy(request, environment_id)
            messages.success(request, _('Deploy started'))
        except Exception:
            msg = _('Unable to deploy. Try again later')
            exceptions.handle(
                request, msg,
                redirect=reverse('horizon:app-catalog:environments:index'))
        redirect_to = reverse('horizon:app-catalog:environments:services',
                              args=(environment_id,))
        return django_http.HttpResponseRedirect(str(redirect_to))


class ShowEnvironmentServices(tables.LinkAction):
    name = 'show'
    verbose_name = _('Manage Components')
    url = 'horizon:app-catalog:environments:services'

    def allowed(self, request, environment):
        return True


class UpdateEnvironmentRow(tables.Row):
    ajax = True

    def __init__(self, table, datum=None):
        super(UpdateEnvironmentRow, self).__init__(table, datum)
        if hasattr(datum, 'status'):
            self.attrs['status'] = datum.status

    def get_data(self, request, environment_id):
        try:
            return api.environment_get(request, environment_id)
        except exc.HTTPNotFound:
            # returning 404 to the ajax call removes the
            # row from the table on the ui
            raise django_http.Http404
        except Exception:
            # let our unified handler take care of errors here
            with api_utils.handled_exceptions(request):
                raise


class UpdateServiceRow(tables.Row):
    ajax = True

    def get_data(self, request, service_id):
        environment_id = self.table.kwargs['environment_id']
        return api.service_get(request, environment_id, service_id)


class UpdateEnvMetadata(tables.LinkAction):
    name = "update_env_metadata"
    verbose_name = _("Update Metadata")
    ajax = False
    icon = "pencil"
    attrs = {"ng-controller": "MetadataModalHelperController as modal"}

    def __init__(self, attrs=None, **kwargs):
        kwargs['preempt'] = True
        self.session_id = None
        super(UpdateEnvMetadata, self).__init__(attrs, **kwargs)

    def get_link_url(self, environment):
        target = json.dumps({
            'environment': environment.id,
            'session': self.session_id
        })
        self.attrs['ng-click'] = (
            "modal.openMetadataModal('muranoenv', %s, true)" % target)
        return "javascript:void(0);"

    def allowed(self, request, environment=None):
        return environment.status not in (consts.STATUS_ID_DEPLOYING,
                                          consts.STATUS_ID_DELETING)

    def update(self, request, datum):
        if datum:
            env_id = datum.id
            self.session_id = api.Session.get_if_available(request, env_id)


class EnvironmentsTable(tables.DataTable):
    name = md_utils.Column(
        'name',
        link='horizon:app-catalog:environments:services',
        verbose_name=_('Name'),
        form_field=forms.CharField(required=False))

    status = tables.Column('status',
                           verbose_name=_('Status'),
                           status=True,
                           status_choices=consts.STATUS_CHOICES,
                           display_choices=consts.STATUS_DISPLAY_CHOICES)

    def get_env_detail_link(self, environment):
        # NOTE: using the policy check for show_environment
        if policy.check((("murano", "show_environment"),),
                        self.request, target={"environment": environment}):
            return reverse("horizon:app-catalog:environments:services",
                           args=(environment.id,))
        return None

    def __init__(self, request, data=None, needs_form_wrapper=None, **kwargs):
        super(EnvironmentsTable,
              self).__init__(request, data=data,
                             needs_form_wrapper=needs_form_wrapper,
                             **kwargs)
        self.columns['name'].get_link_url = self.get_env_detail_link

    class Meta(object):
        name = 'environments'
        verbose_name = _('Environments')
        template = 'environments/_data_table.html'
        row_class = UpdateEnvironmentRow
        status_columns = ['status']
        no_data_message = _('NO ENVIRONMENTS')
        table_actions_menu = (AbandonEnvironment,
                              DeploymentHistory)
        table_actions = (CreateEnvironment, DeployEnvironment,
                         DeleteEnvironment)
        row_actions = (ShowEnvironmentServices, DeployEnvironment,
                       DeleteEnvironment, AbandonEnvironment,
                       UpdateEnvMetadata)


def get_service_details_link(service):
    return reverse('horizon:app-catalog:environments:service_details',
                   args=(service.environment_id, service['?']['id']))


def get_service_type(datum):
    return datum['?'].get(consts.DASHBOARD_ATTRS_KEY, {}).get('name')


class UpdateMetadata(tables.LinkAction):
    name = "update_metadata"
    verbose_name = _("Update Metadata")
    ajax = False
    icon = "pencil"
    attrs = {"ng-controller": "MetadataModalHelperController as modal"}

    def __init__(self, attrs=None, **kwargs):
        kwargs['preempt'] = True
        self.session_id = None
        super(UpdateMetadata, self).__init__(attrs, **kwargs)

    def get_link_url(self, service):
        env_id = self.table.kwargs.get('environment_id')
        comp_id = service['?']['id']
        target = json.dumps({
            'environment': env_id,
            'component': comp_id,
            'session': self.session_id,
        })
        self.attrs['ng-click'] = (
            "modal.openMetadataModal('muranoapp', %s, true)" % target)
        return "javascript:void(0);"

    def allowed(self, request, service=None):
        status, version = _get_environment_status_and_version(request,
                                                              self.table)
        return status != consts.STATUS_ID_DEPLOYING

    def update(self, request, datum):
        env_id = self.table.kwargs.get('environment_id')
        self.session_id = api.Session.get_if_available(request, env_id)


class ServicesTable(tables.DataTable):
    name = md_utils.Column(
        'name',
        verbose_name=_('Name'),
        link=get_service_details_link)

    _type = tables.Column(get_service_type,
                          verbose_name=_('Type'))

    status = tables.Column(lambda datum: datum['?'].get('status'),
                           verbose_name=_('Status'),
                           status=True,
                           status_choices=consts.STATUS_CHOICES,
                           display_choices=consts.STATUS_DISPLAY_CHOICES)
    operation = tables.Column('operation',
                              verbose_name=_('Last operation'),
                              filters=(defaultfilters.urlize, ))
    operation_updated = tables.Column('operation_updated',
                                      verbose_name=_('Time updated'),
                                      filters=(filters.parse_isotime,))

    def get_object_id(self, datum):
        return datum['?']['id']

    def get_apps_list(self):
        packages = []
        with api_utils.handled_exceptions(self.request):
            packages, self._more = pkg_api.package_list(
                self.request,
                filters={'type': 'Application', 'catalog': True})
        return [package.to_dict() for package in packages]

    def actions_allowed(self):
        status, version = _get_environment_status_and_version(
            self.request, self)
        return status not in consts.NO_ACTION_ALLOWED_STATUSES

    def get_categories_list(self):
        return catalog_views.get_categories_list(self.request)

    def get_row_actions(self, datum):
        actions = super(ServicesTable, self).get_row_actions(datum)
        environment_id = self.kwargs['environment_id']
        app_actions = []
        for action_datum in api.extract_actions_list(datum):
            _classes = ('murano_action',)

            class CustomAction(tables.LinkAction):
                name = action_datum['name']
                verbose_name = action_datum.get('title') or name
                url = reverse('horizon:app-catalog:environments:start_action',
                              args=(environment_id, action_datum['id']))
                classes = _classes
                table = self

                def allowed(self, request, datum):
                    status, version = _get_environment_status_and_version(
                        request, self.table)
                    if status in consts.NO_ACTION_ALLOWED_STATUSES:
                        return False
                    return True

            bound_action = CustomAction()
            if not bound_action.allowed(self.request, datum):
                continue
            bound_action.datum = datum
            if issubclass(bound_action.__class__, tables.LinkAction):
                bound_action.bound_url = bound_action.get_link_url(datum)
            app_actions.append(bound_action)
        if app_actions:
            # Show native actions first (such as "Delete Component") and
            # then add sorted application actions
            actions.extend(sorted(app_actions, key=lambda x: x.name))
        return actions

    def get_repo_url(self):
        return pkg_consts.DISPLAY_MURANO_REPO_URL

    def get_pkg_def_url(self):
        return reverse('horizon:app-catalog:packages:index')

    class Meta(object):
        name = 'services'
        verbose_name = _('Component List')
        no_data_message = _('No components')
        status_columns = ['status']
        row_class = UpdateServiceRow
        table_actions = (AddApplication, DeployThisEnvironment)
        row_actions = (DeleteService, UpdateMetadata)
        multi_select = False


class ShowDeploymentDetails(tables.LinkAction):
    name = 'show_deployment_details'
    verbose_name = _('Show Details')

    def get_link_url(self, deployment=None):
        kwargs = {'environment_id': deployment.environment_id,
                  'deployment_id': deployment.id}
        return reverse('horizon:app-catalog:environments:deployment_details',
                       kwargs=kwargs)

    def allowed(self, request, environment):
        return True


class DeploymentsTable(tables.DataTable):
    started = tables.Column('started',
                            verbose_name=_('Time Started'),
                            filters=(filters.parse_isotime,))
    finished = tables.Column('finished',
                             verbose_name=_('Time Finished'),
                             filters=(filters.parse_isotime,))

    status = tables.Column(
        'state',
        verbose_name=_('Status'),
        status=True,
        status_choices=consts.DEPLOYMENT_STATUS_CHOICES,
        display_choices=consts.DEPLOYMENT_STATUS_DISPLAY_CHOICES)

    class Meta(object):
        name = 'deployments'
        verbose_name = _('Deployments')
        row_actions = (ShowDeploymentDetails,)


class EnvConfigTable(tables.DataTable):
    name = md_utils.Column('name', verbose_name=_('Name'))
    _type = tables.Column(
        lambda datum: get_service_type(datum) or 'Unknown',
        verbose_name=_('Type'))

    def get_object_id(self, datum):
        return datum['?']['id']

    class Meta(object):
        name = 'environment_configuration'
        verbose_name = _('Deployed Components')


def get_deployment_history_reports(deployment):
    template_name = 'deployments/_cell_reports.html'
    context = {
        "reports": deployment.reports,
    }
    return template.loader.render_to_string(template_name, context)


def get_deployment_history_services(deployment):
    template_name = 'deployments/_cell_services.html'
    services = {}
    for service in deployment.description['services']:
        service_type = service['?']['type']
        if service_type.find('/') != -1:
            service_type = service_type[:service_type.find('/')]
        services[service.get('name', service['?']['name'])] = service_type
    context = {
        "services": services,
    }
    return template.loader.render_to_string(template_name, context)


class DeploymentHistoryTable(tables.DataTable):
    environment_name = tables.WrappingColumn(
        lambda d: d.description['name'],
        verbose_name=_('Environment'))
    logs = tables.Column(get_deployment_history_reports,
                         verbose_name=_('Logs (Created, Message)'))
    services = tables.Column(get_deployment_history_services,
                             verbose_name=_('Services (Name, Type)'))
    status = tables.Column(
        'state',
        verbose_name=_('Status'),
        status=True,
        status_choices=consts.DEPLOYMENT_STATUS_CHOICES,
        display_choices=consts.DEPLOYMENT_STATUS_DISPLAY_CHOICES)

    class Meta(object):
        name = 'deployment_history'
        verbose_name = _('Deployment History')
        row_actions = (ShowDeploymentDetails,)



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from muranoclient.v1 import client
from muranodashboard.api import packages
from openstack_dashboard.test import helpers


def mock_next(obj, attr, value):
    setattr(obj.__next__, attr, value)


class MagicIterMock(mock.MagicMock):
    __next__ = mock.Mock(return_value=None)


class TestPackagesAPI(helpers.APITestCase):
    def setUp(self):
        super(TestPackagesAPI, self).setUp()

        self.packages = ['foo', 'bar', 'baz']
        self.mock_client = mock.Mock(spec=client)
        self.mock_client.packages.filter.return_value = self.packages

        packages.api = mock.Mock()
        packages.api.muranoclient.return_value = self.mock_client
        self.mock_request = mock.Mock()

        self.addCleanup(mock.patch.stopall)

    def test_package_list(self):
        package_list, more = packages.package_list(self.mock_request)

        self.assertEqual(self.packages, package_list)
        self.assertFalse(more)
        self.mock_client.packages.filter.assert_called_once_with(limit=100)

    def test_package_list_with_paginate(self):
        package_list, more = packages.package_list(self.mock_request,
                                                   paginate=True,
                                                   page_size=1)

        # Only one package should be returned.
        self.assertEqual(self.packages[:1], package_list)
        self.assertTrue(more)
        self.mock_client.packages.filter.assert_called_once_with(limit=2)
        self.mock_client.packages.filter.reset_mock()

        package_list, more = packages.package_list(self.mock_request,
                                                   paginate=True,
                                                   page_size=2)

        # Only two packages should be returned.
        self.assertEqual(self.packages[:2], package_list)
        self.assertTrue(more)
        self.mock_client.packages.filter.assert_called_once_with(limit=3)

    def test_package_list_with_filters(self):
        package_list, more = packages.package_list(self.mock_request,
                                                   marker='test_marker',
                                                   sort_dir='test_sort_dir')

        self.assertEqual(self.packages, package_list)
        self.assertFalse(more)
        self.mock_client.packages.filter.assert_called_once_with(
            limit=100, marker='test_marker', sort_dir='test_sort_dir')

    def test_apps_that_inherit(self):
        setattr(packages.settings, "MURANO_USE_GLARE", False)
        apps = packages.apps_that_inherit(self.mock_request, 'test_fqn')
        self.assertEqual([], apps)

        setattr(packages.settings, "MURANO_USE_GLARE", True)
        apps = packages.apps_that_inherit(self.mock_request, 'test_fqn')
        self.assertEqual(self.packages, apps)

        self.mock_client.packages.filter.assert_called_once_with(
            inherits='test_fqn')

    def test_app_by_fqn(self):
        self.mock_client = MagicIterMock(spec=client)
        mock_next(
            self.mock_client.packages.filter(),
            'return_value',
            self.packages[0]
        )
        packages.api.muranoclient.return_value = self.mock_client
        self.mock_client.reset_mock()

        setattr(packages.settings, "MURANO_USE_GLARE", True)
        app = packages.app_by_fqn(self.mock_request, 'test_fqn', version='1.0')

        self.assertIsNotNone(app)
        self.assertEqual(self.packages[0], app)
        self.mock_client.packages.filter.assert_called_once_with(
            fqn='test_fqn', catalog=True, version='1.0')

    def test_app_by_fqn_except_stop_iteration(self):
        self.mock_client = MagicIterMock(spec=client)
        mock_next(
            self.mock_client.packages.filter(),
            'side_effect',
            StopIteration
        )
        packages.api.muranoclient.return_value = self.mock_client
        self.mock_client.reset_mock()

        setattr(packages.settings, "MURANO_USE_GLARE", True)
        app = packages.app_by_fqn(self.mock_request, 'test_fqn', version='1.0')

        self.assertIsNone(app)
        self.mock_client.packages.filter.assert_called_once_with(
            fqn='test_fqn', catalog=True, version='1.0')

    def test_make_loader_cls(self):
        loader = packages.make_loader_cls()
        self.assertIsNotNone(loader)
        self.assertIn("Loader", str(loader))

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_app_ui(self, *args):
        mock_get_ui = packages.api.muranoclient().packages.get_ui
        mock_get_ui.return_value = 'foo_ui'

        ui = packages.get_app_ui(None, 'foo_app_id')
        mock_args = [arg for arg in mock_get_ui.call_args]

        self.assertEqual(ui, 'foo_ui')
        mock_get_ui.assert_called_once_with('foo_app_id', mock.ANY)
        self.assertEqual('Loader', mock_args[0][1].__name__)

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_app_logo(self, *args):
        mock_get_app_logo = packages.api.muranoclient().packages.get_logo
        mock_get_app_logo.return_value = 'foo_app_logo'

        app_logo = packages.get_app_logo(None, 'foo_app_id')

        self.assertEqual(app_logo, 'foo_app_logo')
        mock_get_app_logo.assert_called_once_with('foo_app_id')

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_app_supplier_logo(self, *args):
        mock_get_supplier_logo = packages.api.muranoclient().packages. \
            get_supplier_logo
        mock_get_supplier_logo.return_value = 'foo_app_supplier_logo'

        app_supplier_logo = packages.get_app_supplier_logo(None, 'foo_app_id')

        self.assertEqual(app_supplier_logo, 'foo_app_supplier_logo')
        mock_get_supplier_logo.assert_called_once_with('foo_app_id')

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_app_fqn(self, *args):
        mock_app = mock.Mock(fully_qualified_name='foo_app_fqn')
        mock_get_app = packages.api.muranoclient().packages.get
        mock_get_app.return_value = mock_app

        app_fqn = packages.get_app_fqn(None, 'foo_app_id')

        self.assertEqual(app_fqn, 'foo_app_fqn')
        mock_get_app.assert_called_once_with('foo_app_id')

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_service_name(self, *args):
        mock_app = mock.Mock()
        mock_app.configure_mock(name='foo_app_name')
        mock_get_app = packages.api.muranoclient().packages.get
        mock_get_app.return_value = mock_app

        app_service_name = packages.get_service_name(None, 'foo_app_id')

        self.assertEqual(app_service_name, 'foo_app_name')
        mock_get_app.assert_called_once_with('foo_app_id')

    @mock.patch('muranodashboard.common.cache._load_from_file',
                return_value=None)
    @mock.patch('muranodashboard.common.cache._save_to_file')
    def test_get_package_details(self, *args):
        mock_app = mock.Mock()
        mock_get_app = packages.api.muranoclient().packages.get
        mock_get_app.return_value = mock_app

        app_details = packages.get_package_details(None, 'foo_app_id')

        self.assertEqual(app_details, mock_app)
        mock_get_app.assert_called_once_with('foo_app_id')



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import unittest
from unittest import mock
from yaql.language import contexts as yaql_contexts

from django import forms as django_forms

from muranodashboard.dynamic_ui import fields
from muranodashboard.dynamic_ui import forms
from muranodashboard.dynamic_ui import yaql_expression


class TestAnyFieldDict(unittest.TestCase):

    def test_missing(self):
        any_field_dict = forms.AnyFieldDict()
        result = any_field_dict.__missing__(['foo', 'bar'])
        self.assertEqual('DynamicSelect', result.__name__)


class TestDynamicUiForm(unittest.TestCase):

    def test_collect_fields_process_widget(self):
        test_spec = {
            'type': 'text',
            'name': 'foo_spec',
            'widget_media': {
                'js': 'foo.js',
                'css': 'foo.css'
            },
            'widget_attrs': {'foo': 'bar'}
        }

        result = forms._collect_fields([test_spec], 'foo_form', None)

        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result))
        self.assertEqual('foo_spec', result[0][0])
        self.assertIsInstance(result[0][1], fields.CharField)

        field = result[0][1]

        self.assertTrue(hasattr(field, 'widget'))
        self.assertTrue(hasattr(field.widget, 'Media'))
        self.assertEqual('foo.js', field.widget.Media.js)
        self.assertEqual('foo.css', field.widget.Media.css)
        self.assertIn('foo', field.widget.__dict__['attrs'])
        self.assertEqual('bar', field.widget.__dict__['attrs']['foo'])

    def test_collect_fields_parse_spec_with_yaql_expression(self):
        mock_yaql = mock.Mock(spec=yaql_expression.YaqlExpression)
        test_spec = {
            'type': 'choice',
            'name': 'foo_spec',
            'validators': [{'expr': mock_yaql}]
        }

        result = forms._collect_fields([test_spec], 'foo_form', None)

        self.assertIsInstance(result, list)
        self.assertEqual(1, len(result))
        self.assertEqual('foo_spec', result[0][0])
        self.assertIsInstance(result[0][1], fields.ChoiceField)

        field = result[0][1]

        self.assertTrue(hasattr(field, 'validators'))
        self.assertIn('expr', field.validators[0])
        self.assertIsInstance(field.validators[0]['expr'], fields.RawProperty)
        self.assertEqual(mock_yaql, field.validators[0]['expr'].spec)


class TestDynamicFormMetaclass(unittest.TestCase):

    def test_new(self):
        test_dict = {
            'name': 'foo_form',
            'field_specs': [{'type': 'text', 'name': 'foo_spec'}],
            'service': 'foo_service',
        }

        form = forms.DynamicFormMetaclass('foo_form', (), test_dict)
        self.assertEqual('foo_form', form.__name__)
        self.assertEqual('foo_service', form.service)
        self.assertIsInstance(form.declared_fields, dict)
        self.assertIn('foo_spec', form.declared_fields)
        self.assertIsInstance(form.declared_fields['foo_spec'],
                              fields.CharField)


class TestUpdatableFieldsForm(unittest.TestCase):

    def setUp(self):
        super(TestUpdatableFieldsForm, self).setUp()

        self.form = forms.UpdatableFieldsForm(None)
        self.assertEqual('required', self.form.required_css_class)

    def test_update_fields(self):
        mock_password_field = mock.Mock(spec=fields.PasswordField)
        mock_password_field.confirm_input = True
        mock_password_field.has_clone = False
        mock_password_field.original = True
        mock_password_field.required = True
        mock_password_field.update = mock.Mock()
        mock_password_field.initial = 'foo_initial'
        mock_password_field.get_clone_name.return_value = 'bar_password_field'
        mock_clone_password_field = mock.Mock(
            spec=fields.PasswordField, required=True)
        mock_password_field.clone_field.return_value = \
            mock_clone_password_field

        self.form.fields = collections.OrderedDict({
            'foo_password_field': mock_password_field})
        self.form.update_fields()

        self.assertEqual(2, len(self.form.fields))
        self.assertIn('foo_password_field', self.form.fields)
        self.assertIn('bar_password_field', self.form.fields)
        self.assertEqual(mock_password_field,
                         self.form.fields['foo_password_field'])
        self.assertEqual(mock_clone_password_field,
                         self.form.fields['bar_password_field'])
        mock_password_field.get_clone_name.assert_called_once_with(
            'foo_password_field')
        self.assertTrue(mock_password_field.get_clone_name.called)
        self.assertTrue(mock_password_field.update.called)


class TestServiceConfigurationForm(unittest.TestCase):

    def setUp(self):
        super(TestServiceConfigurationForm, self).setUp()

        mock_service = mock.Mock()
        mock_service.update_cleaned_data.return_value = \
            {'foo': 'bar', 'baz': 'qux'}

        self.form = forms.ServiceConfigurationForm(
            initial={'app_id': 'foo_app'})
        self.form._errors = []
        self.form.validators = []
        self.form.cleaned_data = {'foo': 'bar', 'baz': 'qux'}
        self.form.service = mock_service

        self.assertEqual('foo_app_%s', self.form.auto_id)
        self.assertIsInstance(self.form.context, yaql_contexts.Context)

    @mock.patch.object(forms, 'LOG', autospec=True)
    def test_clean(self, mock_log):
        password_field = mock.Mock(spec=fields.PasswordField)
        password_field.enabled = True
        password_field.confirm_input = True

        foo_field = mock.Mock()
        foo_field.enabled = False
        foo_field.postclean.return_value = 'post_foo'

        self.form.fields = {'foo': foo_field, 'password': password_field}
        result = self.form.clean()
        expected_cleaned_data = {'foo': 'post_foo', 'baz': 'qux'}

        for key, val in expected_cleaned_data.items():
            self.assertEqual(val, result[key])

        # NOTE(felipemonteiro): mock.ANY is being used in the assertions
        # below, rather than `{'foo': 'bar', 'baz': 'qux'}` because
        # `cleaned_data[name] = value` in clean() appears to also change the
        # dict that was passed in to mock objects in previous lines of code.
        foo_field.postclean.assert_called_once_with(self.form, 'foo', mock.ANY)
        password_field.compare.assert_called_once_with('password', mock.ANY)
        mock_log.debug.assert_called_once_with(
            "Update 'foo' data in postclean method")
        self.form.service.update_cleaned_data.assert_called_with(
            mock.ANY, form=self.form)

    def test_clean_except_validation_error(self):
        mock_expr = mock.Mock()
        mock_expr.evaluate.return_value = False
        test_validator = {'expr': mock_expr, 'message': 'Foo Error'}

        self.form.validators = [test_validator]
        with self.assertRaises(django_forms.ValidationError) as cm:
            self.form.clean()
        e = cm.exception
        self.assertIn('Foo Error', e.messages)

        self.form.service.update_cleaned_data.assert_called_once_with(
            {'foo': 'bar', 'baz': 'qux'}, form=self.form)
        mock_expr.evaluate.assert_called_once_with(
            data={'foo': 'bar', 'baz': 'qux'}, context=self.form.context)

    def test_clean_with_errors(self):
        self.form._errors = ['foo_error']
        self.assertEqual({'foo': 'bar', 'baz': 'qux'}, self.form.clean())



# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

from django.utils.translation import gettext_lazy as _

from muranodashboard.images import forms


class TestImagesForms(unittest.TestCase):
    def setUp(self):
        super(TestImagesForms, self).setUp()
        metadata = '{"title": "title", "type": "type"}'
        self.mock_img = mock.MagicMock(id=12, murano_image_info=metadata)
        self.mock_request = mock.MagicMock()

    @mock.patch.object(forms, 'LOG')
    def test_filter_murano_images(self, mock_log):
        mock_blank_img = \
            mock.MagicMock(id=13, murano_image_info="info")
        images = [mock_blank_img]
        msg = _('Invalid metadata for image: {0}').format(images[0].id)
        self.assertEqual(images,
                         forms.filter_murano_images(images, self.mock_request))
        mock_log.warning.assert_called_once_with(msg)

        images = [self.mock_img]
        self.assertEqual(images, forms.filter_murano_images(images))

        murano_meta = '{"title": "title", "type": "type"}'

        mock_snapshot_img = mock.MagicMock(
            id=14, murano_image_info=murano_meta, image_type='snapshot')
        images = [mock_snapshot_img]
        self.assertEqual([],
                         forms.filter_murano_images(images, self.mock_request))


class TestMarkImageForm(unittest.TestCase):
    def setUp(self):
        super(TestMarkImageForm, self).setUp()
        self.mock_request = mock.MagicMock()
        self.mark_img_form = forms.MarkImageForm(self.mock_request)

    @mock.patch.object(forms, 'glance')
    def test_handle(self, mock_glance_api):
        data = {
            'title': 'title',
            'image': 'id',
            'type': 'type'
        }
        self.mark_img_form.handle(self.mock_request, data)
        self.assertTrue(mock_glance_api.image_update_properties.called)



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

from horizon import exceptions

from muranodashboard.common import net


class TestNet(unittest.TestCase):

    def setUp(self):
        super(TestNet, self).setUp()

        mock_request = mock.Mock()
        mock_request.user.tenant_id = 'foo_tenant_id'
        self.mock_request = mock_request

        mock_env_patcher = mock.patch.object(net, 'env_api', autospec=True)
        mock_env = mock.Mock()
        mock_env.configure_mock(name='foo')
        self.mock_env_api = mock_env_patcher.start()
        self.mock_env_api.environments_list.return_value = [mock_env]

        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(net, 'neutron', autospec=True)
    def test_get_available_networks_with_filter_one(self, mock_neutron):
        foo_mock_network = mock.Mock(router__external=False,
                                     id='foo-network-id',
                                     subnets=[mock.Mock(id='foo-subnet-id')])
        foo_mock_network.configure_mock(name='foo-network')
        bar_mock_network = mock.Mock()  # Will be excluded by test_filter.
        bar_mock_network.configure_mock(name='bar-network')
        mock_neutron.network_list_for_tenant.return_value = [
            foo_mock_network, bar_mock_network
        ]

        test_filter = r'^foo\-[\w]+'
        result = net.get_available_networks(self.mock_request,
                                            filter=test_filter,
                                            murano_networks='include')

        expected_result = [
            (('foo-network-id', 'foo-subnet-id'), "Network of 'foo'"),
            (('foo-network-id', None), "Network of 'foo': random subnet"),
        ]

        self.assertEqual(expected_result, result)
        mock_neutron.network_list_for_tenant.assert_called_once_with(
            self.mock_request, tenant_id='foo_tenant_id')
        self.mock_env_api.environments_list.assert_called_once_with(
            self.mock_request)

    @mock.patch.object(net, 'neutron', autospec=True)
    def test_get_available_networks_with_filter_none(self, mock_neutron):
        foo_mock_network = mock.Mock(router__external=False,
                                     id='foo-network-id',
                                     subnets=[mock.Mock(id='foo-subnet-id')])
        foo_mock_network.configure_mock(name='foo-network')
        bar_mock_subnet = mock.Mock(
            id='bar-subnet-id', name_or_id='bar-subnet', cidr='255.0.0.0')
        bar_mock_network = mock.Mock(router__external=False,
                                     id='bar-network-id',
                                     name_or_id='bar-network',
                                     subnets=[bar_mock_subnet])
        bar_mock_network.configure_mock(name='bar-network')
        mock_neutron.network_list_for_tenant.return_value = [
            foo_mock_network, bar_mock_network
        ]

        test_filter = r'^[\w]+\-[\w]+'
        result = net.get_available_networks(self.mock_request,
                                            filter=test_filter,
                                            murano_networks='include')

        expected_result = [
            (('foo-network-id', 'foo-subnet-id'), "Network of 'foo'"),
            (('foo-network-id', None), "Network of 'foo': random subnet"),
            (('bar-network-id', 'bar-subnet-id'),
                'bar-network: 255.0.0.0 bar-subnet'),
            (('bar-network-id', None), "bar-network: random subnet"),
        ]

        self.assertEqual(expected_result, result)
        mock_neutron.network_list_for_tenant.assert_called_once_with(
            self.mock_request, tenant_id='foo_tenant_id')
        self.mock_env_api.environments_list.assert_called_once_with(
            self.mock_request)

    @mock.patch.object(net, 'neutron', autospec=True)
    def test_get_available_networks(self, mock_neutron):
        foo_subnets = [
            type('%s-subnet' % k, (object, ),
                 {'id': '%s-subnet-id' % k, 'cidr': '255.0.0.0',
                  'name_or_id': '%s-name-or-id' % k})
            for k in ('fake1', 'fake2')]
        bar_subnets = [
            type('fake3-subnet', (object, ),
                 {'id': 'fake3-subnet-id', 'cidr': '255.255.0.0',
                  'name_or_id': 'fake3-name-or-id'})]
        foo_network = type('FooNetwork', (object, ), {
            'router__external': False,
            'id': 'foo-network-id',
            'subnets': foo_subnets,
            'name': 'foo-network-name',
            'name_or_id': 'foo-network-name-or-id',
        })
        bar_network = type('BarNetwork', (object, ), {
            'router__external': False,
            'id': 'bar-network-id',
            'subnets': bar_subnets,
            'name': 'bar-network-name',
            'name_or_id': 'bar-network-name-or-id',
        })
        mock_neutron.network_list_for_tenant.return_value = [
            foo_network, bar_network,
        ]

        result = net.get_available_networks(
            self.mock_request, filter=None, murano_networks='exclude')

        expected_result = [
            ((foo_network.id, foo_subnets[0].id),
             '%s: %s %s' % (
                 foo_network.name_or_id, foo_subnets[0].cidr,
                 foo_subnets[0].name_or_id)),
            ((foo_network.id, foo_subnets[1].id),
             '%s: %s %s' % (
                 foo_network.name_or_id, foo_subnets[1].cidr,
                 foo_subnets[1].name_or_id)),
            ((foo_network.id, None),
             '%s: random subnet' % foo_network.name_or_id),
            ((bar_network.id, bar_subnets[0].id),
             '%s: %s %s' % (
                 bar_network.name_or_id, bar_subnets[0].cidr,
                 bar_subnets[0].name_or_id)),
            ((bar_network.id, None),
             '%s: random subnet' % bar_network.name_or_id),
        ]

        self.assertIsInstance(result, list)
        self.assertEqual(len(expected_result), len(result))
        for choice in expected_result:
            self.assertIn(choice, result)
        mock_neutron.network_list_for_tenant.assert_called_once_with(
            self.mock_request, tenant_id='foo_tenant_id')
        self.mock_env_api.environments_list.assert_called_once_with(
            self.mock_request)

    @mock.patch.object(net, 'LOG', autospec=True)
    @mock.patch.object(net, 'neutron', autospec=True)
    def test_get_available_networks_except_service_catalog_exception(
            self, mock_neutron, mock_log):
        mock_neutron.network_list_for_tenant.side_effect = \
            exceptions.ServiceCatalogException('test_exception')
        result = net.get_available_networks(self.mock_request)

        self.assertEqual([], result)
        mock_log.warning.assert_called_once_with(
            'Neutron not found. Assuming Nova Network usage')
        mock_neutron.network_list_for_tenant.assert_called_once_with(
            self.mock_request, tenant_id='foo_tenant_id')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import logging
import os
import yaml
import zipfile

from oslo_log import log
from selenium.webdriver.support import wait

from muranodashboard.tests.functional.config import config as cfg
from muranodashboard.tests.functional import consts

log = log.getLogger(__name__).logger
log.setLevel(logging.DEBUG)
log.addHandler(logging.StreamHandler())


MANIFEST = {'Format': 'MuranoPL/1.0',
            'Type': 'Application',
            'Description': 'MockApp for webUI tests',
            'Author': 'Mirantis, Inc',
            'UI': 'mock_ui.yaml'}


class ImageException(Exception):
    message = "Image doesn't exist"

    def __init__(self, im_type):
        self._error_string = (self.message + '\nDetails: {0} image is '
                                             'not found,'.format(im_type))

    def __str__(self):
        return self._error_string


def upload_app_package(client, app_name, data, hot=False,
                       package_dir=consts.PackageDir, **manifest_kwargs):
    try:
        if not hot:
            manifest = os.path.join(package_dir, 'manifest.yaml')
            archive = compose_package(app_name, manifest, package_dir,
                                      **manifest_kwargs)
        else:
            manifest = os.path.join(consts.HotPackageDir, 'manifest.yaml')
            archive = compose_package(app_name, manifest,
                                      consts.HotPackageDir, hot=True,
                                      **manifest_kwargs)
        package = client.packages.create(data, {app_name: open(archive, 'rb')})
        return package.id
    finally:
        os.remove(archive)
        os.remove(manifest)


def compose_package(app_name, manifest, package_dir,
                    require=None, archive_dir=None, hot=False, **kwargs):
    """Composes a murano package

    Composes package `app_name` with `manifest` file as a template for the
    manifest and files from `package_dir`.
    Includes `require` section if any in the manifest file.
    Puts the resulting .zip file into `acrhive_dir` if present or in the
    `package_dir`.
    """
    with open(manifest, 'w') as f:
        fqn = 'io.murano.apps.' + app_name
        mfest_copy = MANIFEST.copy()
        mfest_copy['FullName'] = fqn
        mfest_copy['Name'] = app_name
        if hot:
            mfest_copy['Format'] = 'Heat.HOT/1.0'
        else:
            mfest_copy['Format'] = '1.0'
            mfest_copy['Classes'] = {fqn: 'mock_muranopl.yaml'}
        if require:
            mfest_copy['Require'] = require
        if kwargs:
            mfest_copy.update(kwargs)
        f.write(yaml.dump(mfest_copy, default_flow_style=False))

    name = app_name + '.zip'

    if not archive_dir:
        archive_dir = os.path.dirname(os.path.abspath(__file__))
    archive_path = os.path.join(archive_dir, name)

    with zipfile.ZipFile(archive_path, 'w') as zip_file:
        for root, dirs, files in os.walk(package_dir):
            for f in files:
                zip_file.write(
                    os.path.join(root, f),
                    arcname=os.path.join(os.path.relpath(root, package_dir), f)
                )

    return archive_path


def compose_bundle(bundle_path, app_names):
    """Composes a murano bundle. """
    bundle = {'Packages': []}
    for app_name in app_names:
        bundle['Packages'].append({'Name': app_name})
    with open(bundle_path, 'w') as f:
        f.write(json.dumps(bundle))


def glare_enabled():
    return cfg.common.packages_service == "glare"


class wait_for_page_change(object):
    """Waits until current page change in a browser."""
    def __init__(self, driver, timeout=30):
        self.driver = driver
        self.timeout = timeout

    def __enter__(self):
        self.old_page = self.driver.find_element_by_tag_name('html')

    def page_has_loaded(self):
        new_page = self.driver.find_element_by_tag_name('html')
        return new_page.id != self.old_page.id

    def __exit__(self, *_):
        wait.WebDriverWait(self.driver, timeout=self.timeout).until(
            lambda _: self.page_has_loaded()
        )



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os

from oslo_log import log as logging

from muranodashboard.common import utils
from muranodashboard.environments import consts


LOG = logging.getLogger(__name__)
OBJS_PATH = os.path.join(consts.CACHE_DIR, 'apps')

if not os.path.exists(OBJS_PATH):
    os.makedirs(OBJS_PATH)
    LOG.info('Creating apps cache directory located at {dir}'.
             format(dir=OBJS_PATH))

LOG.info('Using apps cache directory located at {dir}'.
         format(dir=OBJS_PATH))


def _get_entry_path(app_id):
    head, tail = app_id[:2], app_id[2:]
    head = os.path.join(OBJS_PATH, head)
    if not os.path.exists(head):
        os.mkdir(head)
    tail = os.path.join(head, tail)
    if not os.path.exists(tail):
        os.mkdir(tail)
    return tail


def _load_from_file(file_name):
    if os.path.isfile(file_name) and os.path.getsize(file_name) > 0:
        with open(file_name, 'rb') as f:
            p = utils.CustomUnpickler(f)
            return p.load()
    return None


def _save_to_file(file_name, content):
    dir_path = os.path.dirname(file_name)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    with open(file_name, 'wb') as f:
        p = utils.CustomPickler(f)
        p.dump(content)


def with_cache(*dst_parts):
    def _decorator(func):
        @functools.wraps(func)
        def __inner(request, app_id):
            path = os.path.join(_get_entry_path(app_id), *dst_parts)
            # Remove file extensions since file content is pickled and
            # could not be open as usual files
            path = os.path.splitext(path)[0] + '-pickled'
            content = _load_from_file(path)
            if content is None:
                content = func(request, app_id)
                if content:
                    LOG.debug('Caching value at {0}.'.format(path))
                    _save_to_file(path, content)
            else:
                LOG.debug('Using cached value from {0}.'.format(path))

            return content

        return __inner

    return _decorator



#    Copyright (c) 2016 AT&T Corp
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from django import forms as django_forms

from muranoclient.common import exceptions as exc
from muranodashboard.packages import consts
from muranodashboard.packages import forms
from openstack_dashboard.test import helpers


class TestImportBundleForm(helpers.APITestCase):

    def test_clean_form(self):
        import_bundle_form = forms.ImportBundleForm()
        expected = {'import_type': 'by_name', 'name': 'test_form_name'}
        import_bundle_form.cleaned_data = expected

        cleaned_data = import_bundle_form.clean()
        self.assertEqual(expected, cleaned_data)

    def test_clean_form_except_validation_error(self):
        import_bundle_form = forms.ImportBundleForm()
        for attr in ['name', 'url']:
            import_bundle_form.cleaned_data = {
                'import_type': 'by_{0}'.format(attr), attr: None}
            expected_error_msg = 'Please supply a bundle {0}'.format(attr)

            with self.assertRaisesRegex(django_forms.ValidationError,
                                        expected_error_msg):
                import_bundle_form.clean()


class TestImportPackageForm(helpers.APITestCase):

    def setUp(self):
        super(TestImportPackageForm, self).setUp()

        self.import_pkg_form = forms.ImportPackageForm()
        fields = self.import_pkg_form.fields
        self.assertEqual(
            'Optional',
            fields['repo_version'].widget.attrs['placeholder'])

    def test_clean_package(self):
        size_in_bytes = (consts.MAX_FILE_SIZE_MB - 1) << 20
        mock_package = mock.Mock(size=size_in_bytes)
        self.import_pkg_form.cleaned_data = {
            'package': mock_package
        }

        cleaned_data = self.import_pkg_form.clean_package()
        self.assertEqual(mock_package, cleaned_data)

    def test_clean_package_exception_validation_error(self):
        size_in_bytes = (consts.MAX_FILE_SIZE_MB + 1) << 20
        mock_package = mock.Mock(size=size_in_bytes)
        self.import_pkg_form.cleaned_data = {
            'package': mock_package
        }

        expected_error_msg = 'It is forbidden to upload files larger than '\
                             '{0} MB.'.format(consts.MAX_FILE_SIZE_MB)
        with self.assertRaisesRegex(django_forms.ValidationError,
                                    expected_error_msg):
            self.import_pkg_form.clean_package()

    def test_clean_form(self):
        expected = {
            'import_type': 'by_name',
            'repo_name': 'test_repo_name'
        }
        self.import_pkg_form.cleaned_data = expected

        cleaned_data = self.import_pkg_form.clean()
        self.assertEqual(expected, cleaned_data)

    def test_clean_form_except_validation_error(self):
        tuples = (
            ('upload', 'package', 'file'),
            ('by_name', 'repo_name', 'name'),
            ('by_url', 'url', 'url')
        )
        for _tuple in tuples:
            self.import_pkg_form.cleaned_data = {
                'import_type': '{0}'.format(_tuple[0]), _tuple[1]: None}
            expected_error_msg = 'Please supply a package {0}'.\
                                 format(_tuple[2])

            with self.assertRaisesRegex(django_forms.ValidationError,
                                        expected_error_msg):
                self.import_pkg_form.clean()


class TestUpdatePackageForm(helpers.APITestCase):

    def setUp(self):
        super(TestUpdatePackageForm, self).setUp()

        mock_package = mock.MagicMock(
            tags=['bar', 'baz', 'qux'], is_public=False,
            enabled=True, description='quux')
        mock_package.configure_mock(name='foo')
        fake_response = {'status_code': 200}
        self.mock_request = mock.MagicMock(return_value=fake_response, META=[])
        kwargs = {'request': self.mock_request, 'package': mock_package}
        self.update_pkg_form = forms.UpdatePackageForm(**kwargs)

    def test_set_initial(self):
        # set_initial was already called by forms.UpdatePackageForm(**kwargs)
        self.assertEqual('foo', self.update_pkg_form.fields['name'].initial)
        self.assertEqual('bar, baz, qux',
                         self.update_pkg_form.fields['tags'].initial)
        self.assertEqual(False,
                         self.update_pkg_form.fields['is_public'].initial)
        self.assertEqual(True, self.update_pkg_form.fields['enabled'].initial)
        self.assertEqual('quux',
                         self.update_pkg_form.fields['description'].initial)


class TestModifyPackageForm(helpers.APITestCase):

    def setUp(self):
        super(TestModifyPackageForm, self).setUp()

        mock_package = mock.MagicMock(
            type='Application', tags=['bar', 'baz', 'qux'], is_public=False,
            enabled=True, description='quux', categories=['c1', 'c2'])
        mock_package.configure_mock(name='foo')
        self.kwargs = {'initial': {'package': mock_package}}

        fake_response = {
            "status_code": 200,
            "text": '{"foo": "bar"}',
        }
        self.mock_request = mock.MagicMock(return_value=(fake_response))

        with mock.patch('muranodashboard.api.muranoclient') as mock_client:
            mock_categories = []
            for cname in ['c3', 'c4']:
                mock_category = mock.Mock()
                mock_category.configure_mock(name=cname)
                mock_categories.append(mock_category)
            mock_client().categories.list.return_value = mock_categories
            self.modify_pkg_form = forms.ModifyPackageForm(self.mock_request,
                                                           **self.kwargs)

    def test_init(self):
        self.assertEqual(
            [('c3', 'c3'), ('c4', 'c4')],
            self.modify_pkg_form.fields['categories'].choices)

        for key in ('c1', 'c2'):
            self.assertIn(key,
                          self.modify_pkg_form.fields['categories'].initial)
            self.assertEqual(
                True, self.modify_pkg_form.fields['categories'].initial[key])

    @mock.patch.object(forms, 'exceptions')
    @mock.patch.object(forms, 'reverse')
    @mock.patch('muranodashboard.api.muranoclient')
    def test_init_except_http_exception(self, mock_client, mock_reverse,
                                        mock_exceptions):
        mock_client().categories.list.side_effect = exc.HTTPException
        mock_reverse.return_value = 'test_redirect'

        self.modify_pkg_form = forms.ModifyPackageForm(self.mock_request,
                                                       **self.kwargs)
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request, 'Unable to get list of categories',
            redirect='test_redirect')

    @mock.patch('muranodashboard.api.muranoclient')
    def test_handle(self, mock_client):
        mock_client().packages.update.return_value = {'status_code': 200}
        test_data = {'tags': 't1 ,t2  ,t3'}
        self.modify_pkg_form.initial['app_id'] = 'test_app_id'

        result = self.modify_pkg_form.handle(self.mock_request, test_data)

        self.assertEqual({'status_code': 200}, result)
        mock_client().packages.update.assert_called_once_with(
            'test_app_id', {'tags': ['t1', 't2', 't3']})

    @mock.patch.object(forms, 'exceptions')
    @mock.patch.object(forms, 'reverse')
    @mock.patch('muranodashboard.api.muranoclient')
    def test_handle_except_http_forbidden(self, mock_client, mock_reverse,
                                          mock_exceptions):
        mock_client().packages.update.side_effect = exc.HTTPForbidden
        mock_reverse.return_value = 'test_redirect'
        test_data = {'tags': 't1 ,t2  ,t3'}
        self.modify_pkg_form.initial['app_id'] = 'test_app_id'

        self.modify_pkg_form.handle(self.mock_request, test_data)

        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request, 'You are not allowed to perform this operation',
            redirect='test_redirect')

    @mock.patch.object(forms, 'exceptions')
    @mock.patch.object(forms, 'reverse')
    @mock.patch('muranodashboard.api.muranoclient')
    def test_handle_except_http_conflict(self, mock_client, mock_reverse,
                                         mock_exceptions):
        mock_client().packages.update.side_effect = exc.HTTPConflict
        mock_reverse.return_value = 'test_redirect'
        test_data = {'tags': 't1 ,t2  ,t3'}
        self.modify_pkg_form.initial['app_id'] = 'test_app_id'

        self.modify_pkg_form.handle(self.mock_request, test_data)

        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request,
            'Package or Class with the same name is already made public',
            redirect='test_redirect')

    @mock.patch.object(forms, 'exceptions')
    @mock.patch.object(forms, 'reverse')
    @mock.patch('muranodashboard.api.muranoclient')
    def test_handle_except_exception(self, mock_client, mock_reverse,
                                     mock_exceptions):
        e = Exception()
        setattr(e, 'details', '{"error": {"message": "test_error_message"}}')
        mock_client().packages.update.side_effect = e
        mock_reverse.return_value = 'test_redirect'
        test_data = {'tags': 't1 ,t2  ,t3'}
        self.modify_pkg_form.initial['app_id'] = 'test_app_id'

        self.modify_pkg_form.handle(self.mock_request, test_data)

        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request,
            'Failed to modify the package.',
            redirect='test_redirect')


class TestSelectCategories(helpers.APITestCase):

    def setUp(self):
        super(TestSelectCategories, self).setUp()

        fake_response = {
            "status_code": 200,
            "text": '{"foo": "bar"}',
        }
        self.mock_request = mock.MagicMock(return_value=(fake_response))
        self.kwargs = {'request': self.mock_request}

        with mock.patch('muranodashboard.api.muranoclient') as mock_client:
            mock_categories = []
            for cname in ['c1', 'c2']:
                mock_category = mock.Mock()
                mock_category.configure_mock(name=cname)
                mock_categories.append(mock_category)
            mock_client().categories.list.return_value = mock_categories
            self.select_categories_form = forms.SelectCategories(**self.kwargs)

    def test_init(self):
        self.assertEqual(
            [('c1', 'c1'), ('c2', 'c2')],
            self.select_categories_form.fields['categories'].choices)

    @mock.patch.object(forms, 'exceptions')
    @mock.patch.object(forms, 'reverse')
    @mock.patch('muranodashboard.api.muranoclient')
    def test_init_except_http_exception(self, mock_client, mock_reverse,
                                        mock_exceptions):
        mock_client().categories.list.side_effect = exc.HTTPException
        mock_reverse.return_value = 'test_redirect'

        forms.SelectCategories(**self.kwargs)

        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request, 'Unable to get list of categories',
            redirect='test_redirect')



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy
import json
import re

from django.core import validators as django_validator
from django import forms
from django.forms import widgets
from django.template import defaultfilters
from django.urls import reverse
from django.utils.encoding import force_str
from django.utils import html
from django.utils.translation import gettext_lazy as _
from horizon import exceptions
from horizon import forms as hz_forms
from horizon import messages
try:
    from openstack_dashboard.api import _nova as nova
except ImportError:
    from openstack_dashboard.api import nova
from openstack_dashboard.api import cinder
from openstack_dashboard.api import glance
from openstack_dashboard.api import neutron
from oslo_log import log as logging
from oslo_log import versionutils
from yaql import legacy

from muranodashboard.api import packages as pkg_api
from muranodashboard.common import designate
from muranodashboard.common import net
from muranodashboard.dynamic_ui import helpers
from muranodashboard.environments import api as env_api


LOG = logging.getLogger(__name__)


def with_request(func):
    """Injects request into func

    The decorator is meant to be used together with `UpdatableFieldsForm':
    apply it to the `update' method of fields inside that form.
    """
    def update(self, initial, request=None, **kwargs):
        initial_request = initial.get('request')
        for key, value in iter(initial.items()):
            if key != 'request' and key not in kwargs:
                kwargs[key] = value

        if initial_request:
            LOG.debug("Using 'request' value from initial dictionary")
            func(self, initial_request, **kwargs)
        elif request:
            LOG.debug("Using direct 'request' value")
            func(self, request, **kwargs)
        else:
            LOG.error("No 'request' value passed neither via initial "
                      "dictionary, nor directly")
            raise forms.ValidationError("Can't get a request information")
    return update


def make_yaql_validator(validator_property):
    """Field-level validator uses field's value as its '$' root object."""
    expr = validator_property['expr'].spec
    message = validator_property.get('message', '')

    def validator_func(value):
        context = legacy.create_context()
        context['$'] = value
        if not expr.evaluate(context=context):
            raise forms.ValidationError(message)

    return validator_func


def get_regex_validator(expr):
    try:
        validator = expr['validators'][0]
        if isinstance(validator, django_validator.RegexValidator):
            return validator
    except (TypeError, KeyError, IndexError):
        pass
    return None


# This function is needed if we don't want to change existing services
# regexpValidators
def wrap_regex_validator(validator, message):
    def _validator(value):
        try:
            validator(value)
        except forms.ValidationError:
            # provide our own message
            raise forms.ValidationError(message)
    return _validator


def get_murano_images(request, region=None):
    images = []
    try:
        # https://bugs.launchpad.net/murano/+bug/1339261 - glance
        # client version change alters the API. Other tuple values
        # are _more and _prev (in recent glance client)
        with helpers.current_region(request, region):
            images = glance.image_list_detailed(request)[0]
    except Exception:
        LOG.error("Error to request image list from glance ")
        exceptions.handle(request, _("Unable to retrieve public images."))
    murano_images = []
    # filter out the snapshot image type
    images = filter(
        lambda x: x.properties.get("image_type", '') != 'snapshot', images)
    for image in images:
        # Additional properties, whose value is always a string data type, are
        # only included in the response if they have a value.
        murano_property = getattr(image, 'murano_image_info', None)
        if murano_property:
            try:
                murano_metadata = json.loads(murano_property)
            except ValueError:
                LOG.warning("JSON in image metadata is not valid. "
                            "Check it in glance.")
                messages.error(request, _("Invalid murano image metadata"))
            else:
                image.murano_property = murano_metadata
                murano_images.append(image)
    return murano_images


class RawProperty(object):
    def __init__(self, key, spec):
        self.key = key
        self.spec = spec
        self.value = None
        self.value_evaluated = False

    def finalize(self, form_name, service, cls):
        def _get(field):
            if self.value_evaluated:
                return self.value
            return service.get_data(form_name, self.spec)

        def _set(field, value):
            self.value = value
            self.value_evaluated = value is not None
            if hasattr(cls, self.key):
                getattr(cls, self.key).fset(field, value)

        def _del(field):
            _set(field, None)
        return property(_get, _set, _del)


FIELD_ARGS_TO_ESCAPE = ['help_text', 'initial', 'description', 'label']


class CustomPropertiesField(forms.Field):
    js_validation = False

    def __init__(self, description=None, description_title=None,
                 *args, **kwargs):
        self.description = description
        self.description_title = (description_title or
                                  force_str(kwargs.get('label', '')))

        for arg in FIELD_ARGS_TO_ESCAPE:
            if kwargs.get(arg):
                kwargs[arg] = html.escape(force_str(kwargs[arg]))

        validators = []
        validators_js = []
        for validator in kwargs.get('validators', []):
            if hasattr(validator, '__call__'):  # single regexpValidator
                validators.append(validator)
                if hasattr(validator, 'regex'):
                    regex_message = ''
                    error_messages = kwargs.get('error_messages', {})
                    if hasattr(validator, 'code') and \
                       validator.code in error_messages:
                        regex_message = force_str(
                            error_messages[validator.code]
                        )
                    validators_js. \
                        append({'regex': force_str(validator.regex.pattern),
                                'message': regex_message})
            else:  # mixed list of regexpValidator and YAQL validators
                expr = validator.get('expr')
                regex_validator = get_regex_validator(expr)
                regex_message = validator.get('message', '')
                if regex_validator:
                    validators.append(wrap_regex_validator(
                        regex_validator, regex_message))
                elif isinstance(expr, RawProperty):
                    validators.append(validator)
                if hasattr(regex_validator, 'regex'):
                    validators_js.\
                        append({'regex': regex_validator.regex.pattern,
                                'message': regex_message})
        kwargs['validators'] = validators
        if validators_js:
            self.js_validation = json.dumps(validators_js)

        super(CustomPropertiesField, self).__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = super(CustomPropertiesField, self).widget_attrs(widget)
        if self.js_validation:
            attrs['data-validators'] = self.js_validation
        return attrs

    def clean(self, value):
        """Skip all validators if field is disabled."""
        # form is assigned in ServiceConfigurationForm.finalize_fields()
        form = self.form
        # the only place to ensure that Service object has up-to-date
        # cleaned_data
        form.service.update_cleaned_data(form.cleaned_data, form=form)
        if getattr(self, 'enabled', True):
            return super(CustomPropertiesField, self).clean(value)
        else:
            return super(CustomPropertiesField, self).to_python(value)

    @classmethod
    def finalize_properties(cls, kwargs, form_name, service):
        props = {}
        kwargs_ = copy.copy(kwargs)
        for key, value in kwargs_.items():
            if isinstance(value, RawProperty):
                props[key] = value.finalize(form_name, service, cls)
                del kwargs[key]
        if props:
            return type(cls.__name__, (cls,), props)
        else:
            return cls


class CharField(forms.CharField, CustomPropertiesField):
    pass


class PasswordField(CharField):
    special_characters = '!@#$%^&*()_+|\\/.,~?><:{}-'
    password_re = re.compile(r'^.*(?=.*\d)(?=.*[a-z])(?=.*[A-Z])(?=.*[%s]).*$'
                             % special_characters)
    has_clone = False
    original = True
    attrs = {'data-type': 'password'}
    validate_password = django_validator.RegexValidator(
        password_re, _('The password must contain at least one letter, one   \
                               number and one special character'), 'invalid')

    @staticmethod
    def get_clone_name(name):
        return name + '-clone'

    def compare(self, name, form_data):
        if self.original and self.required:
            # run compare only for original fields
            # do not run compare for hidden fields (they are not required)
            if form_data.get(name) != form_data.get(self.get_clone_name(name)):
                raise forms.ValidationError(_(u"{0}{1} don't match").format(
                    self.label, defaultfilters.pluralize(2)))

    def __init__(self, label, *args, **kwargs):
        self.confirm_input = kwargs.pop('confirm_input', True)

        kwargs.update({'label': label,
                       'error_messages': kwargs.get('error_messages', {}),
                       'widget': forms.PasswordInput(attrs=self.attrs,
                                                     render_value=True)})

        validators = kwargs.get('validators')
        help_text = kwargs.get('help_text')

        if not validators:
            # No custom validators, using default validator
            validators = [self.validate_password]
            if not help_text:
                help_text = _(
                    'Enter a complex password with at least one letter, '
                    'one number and one special character')

            kwargs['error_messages'].setdefault(
                'invalid', self.validate_password.message)
            kwargs['min_length'] = kwargs.get('min_length', 7)
            kwargs['max_length'] = kwargs.get('max_length', 255)
            kwargs['widget'] = forms.PasswordInput(attrs=self.attrs,
                                                   render_value=True)
        else:
            if not help_text:
                # NOTE(kzaitsev) There are custom validators for password,
                # but no help text let's leave only a generic message,
                # since we do not know exact constraints
                help_text = _('Enter a password')

        kwargs.update({'validators': validators,
                       'help_text': help_text})

        super(PasswordField, self).__init__(*args, **kwargs)

    def __deepcopy__(self, memo):
        result = super(PasswordField, self).__deepcopy__(memo)
        result.error_messages = copy.deepcopy(self.error_messages)
        return result

    def clone_field(self):
        self.has_clone = True

        field = copy.deepcopy(self)
        field.original = False
        field.label = _('Confirm password')
        field.error_messages['required'] = _('Please confirm your password')
        field.help_text = _('Retype your password')
        return field


class IntegerField(forms.IntegerField, CustomPropertiesField):
    pass


def _get_title(data):
    if isinstance(data, Choice):
        return data.title
    return data


def _disable_non_ready(data):
    if getattr(data, 'enabled', True):
        return {}
    else:
        return {'disabled': 'disabled'}


class ChoiceField(forms.ChoiceField, CustomPropertiesField):
    def __init__(self, **kwargs):
        choices = kwargs.get('choices') or getattr(self, 'choices', None)
        if choices:
            if isinstance(choices, dict):
                choices = list(choices.items())
            kwargs['choices'] = choices
        super(ChoiceField, self).__init__(**kwargs)


class DynamicChoiceField(hz_forms.DynamicChoiceField, CustomPropertiesField):
    pass


class FlavorWidget(widgets.Select):
    def __init__(self, *args, **kwargs):
        super(FlavorWidget, self).__init__(*args, **kwargs)
        self.attrs['class'] = self.attrs.get('class', '') + ' flavor'
        self.attrs['id'] = 'id_flavor'


class FlavorChoiceField(ChoiceField):
    widget = FlavorWidget

    def __init__(self, *args, **kwargs):
        if 'requirements' in kwargs:
            self.requirements = kwargs.pop('requirements')
        super(FlavorChoiceField, self).__init__(*args, **kwargs)

    @with_request
    def update(self, request, form=None, **kwargs):
        choices = []
        with helpers.current_region(request,
                                    getattr(form, 'region', None)):
            flavors = nova.novaclient(request).flavors.list()

        # If no requirements are present, return all the flavors.
        if not hasattr(self, 'requirements'):
            choices = [(flavor.id, flavor.name) for flavor in flavors]
        else:
            for flavor in flavors:
                # If a flavor doesn't meet a minimum requirement,
                # do not add it to the options list and skip to the
                # next flavor.
                if flavor.vcpus < self.requirements.get('min_vcpus', 0):
                    continue
                if flavor.disk < self.requirements.get('min_disk', 0):
                    continue
                if flavor.ram < self.requirements.get('min_memory_mb', 0):
                    continue
                if 'max_vcpus' in self.requirements:
                    if flavor.vcpus > self.requirements['max_vcpus']:
                        continue
                if 'max_disk' in self.requirements:
                    if flavor.disk > self.requirements['max_disk']:
                        continue
                if 'max_memory_mb' in self.requirements:
                    if flavor.ram > self.requirements['max_memory_mb']:
                        continue
                choices.append((flavor.id, flavor.name))

        choices.sort(key=lambda e: e[1])
        self.choices = choices
        if kwargs.get('form'):
            kwargs_form_flavor = kwargs["form"].fields.get('flavor')
        else:
            kwargs_form_flavor = None
        if kwargs_form_flavor:
            self.initial = kwargs["form"]["flavor"].value()
        else:
            # Search through selected flavors
            for flavor_id, flavor_name in self.choices:
                if 'medium' in flavor_name:
                    self.initial = flavor_id
                    break

    def clean(self, value):
        for flavor_id, flavor_name in self.choices:
            if flavor_id == value:
                return flavor_name
        return value


class KeyPairChoiceField(DynamicChoiceField):
    """This widget allows to select keypair for VMs"""
    @with_request
    def update(self, request, form=None, **kwargs):
        self.choices = [('', _('No keypair'))]
        with helpers.current_region(request,
                                    getattr(form, 'region', None)):
            keypairs = nova.novaclient(request).keypairs.list()
        for keypair in sorted(keypairs, key=lambda e: e.name):
            self.choices.append((keypair.name, keypair.name))

        if len(keypairs) == 1:
            self.initial = keypairs[0].name


class SecurityGroupChoiceField(DynamicChoiceField):
    """This widget allows to select a security group for VMs"""
    @with_request
    def update(self, request, **kwargs):
        self.choices = [('', _('Application default security group'))]
        # TODO(pbourke): remove sorted when supported natively in Horizon
        # (https://bugs.launchpad.net/horizon/+bug/1692972)
        for secgroup in sorted(
                neutron.security_group_list(request),
                key=lambda e: e.name_or_id):
            if not secgroup.name_or_id.startswith('murano--'):
                self.choices.append((secgroup.name_or_id, secgroup.name_or_id))


# NOTE(kzaitsev): for transform to work correctly on horizon SelectWidget
# Choice has to be non-string
class Choice(object):
    """A choice that allows disabling specific choices in a SelectWidget."""
    def __init__(self, title, enabled):
        self.title = title
        self.enabled = enabled


class ImageChoiceField(ChoiceField):
    widget = hz_forms.SelectWidget(transform=_get_title,
                                   transform_html_attrs=_disable_non_ready)

    def __init__(self, *args, **kwargs):
        self.image_type = kwargs.pop('image_type', None)
        super(ImageChoiceField, self).__init__(*args, **kwargs)

    @with_request
    def update(self, request, form=None, **kwargs):
        image_map, image_choices = {}, []
        murano_images = get_murano_images(
            request, getattr(form, 'region', None))
        for image in murano_images:
            murano_data = image.murano_property
            title = murano_data.get('title', image.name)

            if image.status == 'active':
                title = Choice(title, enabled=True)
            else:
                title = Choice("{} ({})".format(title, image.status),
                               enabled=False)
            if self.image_type is not None:
                itype = murano_data.get('type')

                if not self.image_type and itype is None:
                    continue

                prefix = '{type}.'.format(type=self.image_type)
                if (not itype.startswith(prefix) and
                        not self.image_type == itype):
                    continue
            image_map[image.id] = title

        for id_, title in sorted(iter(image_map.items()),
                                 key=lambda e: e[1].title):
            image_choices.append((id_, title))

        if image_choices:
            # If only one choice, select it as only option
            if len(image_choices) != 1:
                image_choices.insert(0, ("", _("Select Image")))
        else:
            image_choices.insert(0, ("", _("No images available")))

        self.choices = image_choices


class NetworkChoiceField(ChoiceField):
    def __init__(self,
                 filter=None,
                 murano_networks=None,
                 allow_auto=True,
                 *args,
                 **kwargs):
        self.filter = filter
        if murano_networks:
            if murano_networks.lower() not in ["exclude", "translate"]:
                raise ValueError(_("Invalid value of 'murano_nets' option"))
        self.murano_networks = murano_networks
        self.allow_auto = allow_auto
        super(NetworkChoiceField, self).__init__(*args,
                                                 **kwargs)

    @with_request
    def update(self, request, **kwargs):
        """Populates available networks in the control

        This method is called automatically when the form which contains it is
        rendered
        """
        network_choices = net.get_available_networks(request,
                                                     self.filter,
                                                     self.murano_networks)
        if self.allow_auto:
            network_choices.insert(0, ((None, None), _('Auto')))
        self.choices = network_choices or []

    def to_python(self, value):
        """Converts string representation of widget to tuple value

        Is called implicitly during form cleanup phase
        """
        if value:
            return ast.literal_eval(value)
        else:  # may happen if no networks are available and "Auto" is disabled
            return None, None


class AZoneChoiceField(ChoiceField):
    @with_request
    def update(self, request, form=None, **kwargs):
        try:
            with helpers.current_region(request,
                                        getattr(form, 'region', None)):
                availability_zones = nova.novaclient(
                    request).availability_zones.list(detailed=False)
        except Exception:
            availability_zones = []
            exceptions.handle(request,
                              _("Unable to retrieve  availability zones."))

        az_choices = [(az.zoneName, az.zoneName)
                      for az in availability_zones
                      if az.zoneState['available']]
        if not az_choices:
            az_choices.insert(0, ("", _("No availability zones available")))

        az_choices.sort(key=lambda e: e[1])
        self.choices = az_choices


class VolumeChoiceField(ChoiceField):
    def __init__(self,
                 include_volumes=True,
                 include_snapshots=True,
                 *args,
                 **kwargs):
        self.include_volumes = include_volumes
        self.include_snapshots = include_snapshots
        super(VolumeChoiceField, self).__init__(*args, **kwargs)

    @with_request
    def update(self, request, **kwargs):
        """This widget allows selection of Volumes and Volume Snapshots"""
        available = {'status': cinder.VOLUME_STATE_AVAILABLE}
        choices = []

        if self.include_volumes:
            try:
                choices.extend((volume.id, volume.name)
                               for volume in cinder.volume_list(request,
                               search_opts=available))
            except Exception:
                exceptions.handle(request,
                                  _("Unable to retrieve volume list."))

        if self.include_snapshots:
            try:
                choices.extend((snap.id, snap.name)
                               for snap in cinder.volume_snapshot_list(request,
                               search_opts=available))
            except Exception:
                exceptions.handle(request,
                                  _("Unable to retrieve snapshot list."))

        if choices:
            choices.sort(key=lambda e: e[1])
            choices.insert(0, ("", _("Select volume")))
        else:
            choices.insert(0, ("", _("No volumes available")))
        self.choices = choices


class BooleanField(forms.BooleanField, CustomPropertiesField):
    def __init__(self, *args, **kwargs):
        if 'widget' in kwargs:
            widget = kwargs['widget']
            if isinstance(widget, type):
                widget = widget(attrs={'class': 'checkbox'})
        else:
            widget = forms.CheckboxInput(attrs={'class': 'checkbox'})
        kwargs['widget'] = widget
        kwargs['required'] = False
        super(BooleanField, self).__init__(*args, **kwargs)


@versionutils.deprecated(
    as_of=versionutils.deprecated.JUNO,
    in_favor_of='type boolean (regular BooleanField)',
    remove_in=1)
class FloatingIpBooleanField(BooleanField):
    pass


class ClusterIPField(forms.GenericIPAddressField, CustomPropertiesField):
    def __init__(self, *args, **kwargs):
        super(ClusterIPField, self).__init__(protocol='ipv4', *args, **kwargs)


class DatabaseListField(CharField):
    validate_mssql_identifier = django_validator.RegexValidator(
        re.compile(r'^[a-zA-z_][a-zA-Z0-9_$#@]*$'),
        _(u'First symbol should be latin letter or underscore. Subsequent '
          u'symbols can be latin letter, numeric, underscore, at sign, '
          u'number sign or dollar sign'))

    default_error_messages = {'invalid': validate_mssql_identifier.message}

    def to_python(self, value):
        """Normalize data to a list of strings."""
        if not value:
            return []
        return [name.strip() for name in value.split(',')]

    def validate(self, value):
        """Check if value consists only of valid names."""
        super(DatabaseListField, self).validate(value)
        for db_name in value:
            self.validate_mssql_identifier(db_name)


class ErrorWidget(widgets.Widget):

    def __init__(self, *args, **kwargs):
        self.message = kwargs.pop(
            'message', _("There was an error initialising this field."))
        super(ErrorWidget, self).__init__(*args, **kwargs)

    def render(self, name, value, attrs=None):
        return "<div name={name}>{message}</div>".format(
            name=name, message=self.message)


class MuranoTypeWidget(hz_forms.fields.DynamicSelectWidget):
    def __init__(self, attrs=None, **kwargs):
        if attrs is None:
            attrs = {'class': 'murano_add_select'}
        else:
            attrs.setdefault('class', '')
            attrs['class'] += ' murano_add_select'
        super(MuranoTypeWidget, self).__init__(attrs=attrs, **kwargs)

    class Media(object):
        js = ('muranodashboard/js/add-select.js',)


def make_select_cls(fqns):
    if not isinstance(fqns, (tuple, list)):
        fqns = (fqns,)

    class DynamicSelect(hz_forms.DynamicChoiceField, CustomPropertiesField):
        widget = MuranoTypeWidget

        def __init__(self, empty_value_message=None, *args, **kwargs):
            super(DynamicSelect, self).__init__(*args, **kwargs)
            if empty_value_message is not None:
                self.empty_value_message = empty_value_message
            else:
                self.empty_value_message = _('Select Application')

        @with_request
        def update(self, request, environment_id, **kwargs):
            matching_classes = []
            fqns_seen = set()
            # NOTE(kzaitsev): it's possible to have a private
            # and public apps with the same fqn, however the engine would
            # currently favor private package. Therefore we should squash
            # these until we devise a better way to work with this
            # situation and versioning

            for class_fqn in fqns:
                app_found = pkg_api.app_by_fqn(request, class_fqn)
                if app_found:
                    fqns_seen.add(app_found.fully_qualified_name)
                    matching_classes.append(app_found)

                apps_found = pkg_api.apps_that_inherit(request, class_fqn)
                for app in apps_found:
                    if app.fully_qualified_name in fqns_seen:
                        continue
                    fqns_seen.add(app.fully_qualified_name)
                    matching_classes.append(app)

            if not matching_classes:
                msg = _(
                    "Couldn't find any apps, required for this field.\n"
                    "Tried: {fqns}").format(fqns=', '.join(fqns))
                self.widget = ErrorWidget(message=msg)

            # NOTE(kzaitsev): this closure is needed to allow us have custom
            # logic when clicking add button
            def _make_link():
                ns_url = 'horizon:app-catalog:catalog:add'
                ns_url_args = (environment_id, False, True)

                # This will prevent horizon from adding an extra '+' button
                if not matching_classes:
                    return ''

                return json.dumps([
                    (app.name, reverse(ns_url, args=((app.id,) + ns_url_args)))
                    for app in matching_classes])

            self.widget.add_item_link = _make_link

            apps = env_api.service_list_by_fqns(
                request, environment_id,
                [app.fully_qualified_name for app in matching_classes])
            choices = [('', self.empty_value_message)]
            choices.extend([(app['?']['id'],
                             html.escape(app.name)) for app in apps])
            self.choices = choices
            # NOTE(tsufiev): streamline the drop-down UX: auto-select the
            # single available option in a drop-down
            if len(choices) == 2:
                self.initial = choices[1][0]

        def clean(self, value):
            value = super(DynamicSelect, self).clean(value)
            return None if value == '' else value

    return DynamicSelect


@versionutils.deprecated(
    as_of=versionutils.deprecated.JUNO,
    in_favor_of='type io.murano.windows.ActiveDirectory with a custom '
                'emptyValueMessage attribute',
    remove_in=1)
class DomainChoiceField(make_select_cls('io.murano.windows.ActiveDirectory')):
    def __init__(self, *args, **kwargs):
        super(DomainChoiceField, self).__init__(*args, **kwargs)
        self.choices = [('', _('Not in domain'))]


class ZoneChoiceField(ChoiceField):
    @with_request
    def update(self, request, form=None, **kwargs):
        try:
            zones = designate.zone_list(request)
        except Exception:
            zones = []
            exceptions.handle(request, _("Unable to retrieve DNS zones."))

        # Use zone name instead of ID purely for convenience of being able to
        # pass it through to construct fqdn for apps
        zone_choices = [(z['name'], z['name']) for z in zones]
        zone_choices.sort(key=lambda e: e[1])

        if zones:
            zone_choices.insert(0, ('', _('No DNS zone')))
        else:
            zone_choices.insert(0, ('', _('No DNS zones available')))

        self.choices = zone_choices



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.conf import urls

from muranodashboard.packages import views


urlpatterns = [
    urls.url(r'^$', views.PackageDefinitionsView.as_view(), name='index'),
    urls.url(r'^upload$', views.ImportPackageWizard.as_view(
             views.FORMS), name='upload'),
    urls.url(r'^import_bundle$', views.ImportBundleWizard.as_view(
             views.BUNDLE_FORMS), name='import_bundle'),
    urls.url(r'^modify/(?P<app_id>[^/]+)?$',
             views.ModifyPackageView.as_view(), name='modify'),
    urls.url(r'^(?P<app_id>[^/]+)?$',
             views.DetailView.as_view(), name='detail'),
    urls.url(r'^download/(?P<app_name>[^/]+)/(?P<app_id>[^/]+)?$',
             views.download_packge, name='download'),
]



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import semantic_version

LATEST_FORMAT_VERSION = '2.4'


def check_version(version):
    latest = get_latest_version()
    supported = semantic_version.Version(str(latest.major), partial=True)
    requested = semantic_version.Version.coerce(str(version))
    if supported.major != requested.major:
        msg = 'Unsupported Dynamic UI format version: ' \
              'requested format version {0} is not compatible with the ' \
              'supported family {1}'
        raise ValueError(msg.format(requested, supported))
    if requested > latest:
        msg = 'Unsupported Dynamic UI format version: ' \
              'requested format version {0} is newer than ' \
              'latest supported {1}'
        raise ValueError(msg.format(requested, latest))


def get_latest_version():
    return semantic_version.Version.coerce(LATEST_FORMAT_VERSION)



from django.utils.translation import gettext_lazy as _

# The name of the panel group to be added to HORIZON_CONFIG. Required.
PANEL_GROUP = 'app-catalog_browse_group'
# The display name of the PANEL_GROUP. Required.
PANEL_GROUP_NAME = _('Browse')
# The name of the dashboard the PANEL_GROUP associated with. Required.
PANEL_GROUP_DASHBOARD = 'app-catalog'
